@@ -45,6 +45,10 @@ using namespace std;
 
 #include "auth/KeyRing.h"
 
+#include "include/assert.h"
+
+#define dout_subsys ceph_subsys_mds
+
 void usage()
 {
   derr << "usage: ceph-mds -i name [flags] [[--journal_check rank]|[--hot-standby][rank]]\n"
@@ -70,7 +74,7 @@ static int do_cmds_special_action(const std::string &action,
 {
   common_init_finish(g_ceph_context);
   Messenger *messenger = Messenger::create(g_ceph_context,
-					   entity_name_t::CLIENT(),
+					   entity_name_t::CLIENT(), "mds",
 					   getpid());
   int r = messenger->bind(g_conf->public_addr);
   if (r < 0)
@@ -229,10 +233,8 @@ int main(int argc, const char **argv)
     usage();
   }
 
-  global_print_banner();
-
   Messenger *messenger = Messenger::create(g_ceph_context,
-					   entity_name_t::MDS(-1),
+					   entity_name_t::MDS(-1), "mds",
 					   getpid());
   messenger->set_cluster_protocol(CEPH_MDS_PROTOCOL);
 
@@ -245,11 +247,11 @@ int main(int argc, const char **argv)
     CEPH_FEATURE_PGID64;
   uint64_t required =
     CEPH_FEATURE_OSDREPLYMUX;
-  messenger->set_default_policy(Messenger::Policy::client(supported, required));
+  messenger->set_default_policy(Messenger::Policy::lossy_client(supported, required));
   messenger->set_policy(entity_name_t::TYPE_MON,
-			Messenger::Policy::client(supported,
-						  CEPH_FEATURE_UID |
-						  CEPH_FEATURE_PGID64));
+			Messenger::Policy::lossy_client(supported,
+							CEPH_FEATURE_UID |
+							CEPH_FEATURE_PGID64));
   messenger->set_policy(entity_name_t::TYPE_MDS,
 			Messenger::Policy::lossless_peer(supported,
 							 CEPH_FEATURE_UID));
