@@ -41,6 +41,8 @@ using namespace std;
 #include "global/global_init.h"
 #include "global/signal_handler.h"
 
+#include "include/assert.h"
+
 #define dout_subsys ceph_subsys_mon
 
 extern CompatSet get_ceph_mon_feature_compat_set();
@@ -104,6 +106,11 @@ int main(int argc, const char **argv)
     usage();
   }
 
+  if (g_conf->name.get_id().empty()) {
+    cerr << "must specify id (--id <id> or --name mon.<id>)" << std::endl;
+    usage();
+  }
+
   // -- mkfs --
   if (mkfs) {
     // resolve public_network -> public_addr
@@ -124,17 +131,18 @@ int main(int argc, const char **argv)
       }
       try {
 	monmap.decode(monmapbl);
+
+	// always mark seed/mkfs monmap as epoch 0
+	monmap.set_epoch(0);
       }
       catch (const buffer::error& e) {
 	cerr << argv[0] << ": error decoding monmap " << g_conf->monmap << ": " << e.what() << std::endl;
 	exit(1);
       }      
     } else {
-      int err = MonClient::build_initial_monmap(g_ceph_context, monmap);
+      int err = monmap.build_initial(g_ceph_context, cerr);
       if (err < 0) {
-	cerr << argv[0] << ": error generating initial monmap: " << cpp_strerror(err) << std::endl;
-	usage();
-	exit(1);
+	cerr << argv[0] << ": warning: no initial monitors; must use admin socket to feed hints" << std::endl;
       }
 
       // am i part of the initial quorum?
@@ -278,7 +286,7 @@ int main(int argc, const char **argv)
       tmp.set_epoch(v);
     }
     bufferlist mapbl;
-    tmp.encode(mapbl);
+    tmp.encode(mapbl, CEPH_FEATURES_ALL);
     bufferlist final;
     ::encode(v, final);
     ::encode(mapbl, final);
@@ -346,7 +354,7 @@ int main(int argc, const char **argv)
       ipaddr = g_conf->public_addr;
     } else {
       MonMap tmpmap;
-      int err = MonClient::build_initial_monmap(g_ceph_context, tmpmap);
+      int err = tmpmap.build_initial(g_ceph_context, cerr);
       if (err < 0) {
 	cerr << argv[0] << ": error generating initial monmap: " << cpp_strerror(err) << std::endl;
 	usage();
@@ -378,15 +386,31 @@ int main(int argc, const char **argv)
     CEPH_FEATURE_PGID64;
   messenger->set_default_policy(Messenger::Policy::stateless_server(supported, 0));
   messenger->set_policy(entity_name_t::TYPE_MON,
-                        Messenger::Policy::lossless_peer(supported,
-                                                         CEPH_FEATURE_UID |
-                                                         CEPH_FEATURE_PGID64));
+                        Messenger::Policy::lossless_peer_reuse(supported,
+							       CEPH_FEATURE_UID |
+							       CEPH_FEATURE_PGID64));
   messenger->set_policy(entity_name_t::TYPE_OSD,
                         Messenger::Policy::stateless_server(supported,
                                                             CEPH_FEATURE_PGID64 |
                                                             CEPH_FEATURE_OSDENC));
+  messenger->set_policy(entity_name_t::TYPE_CLIENT,
+			Messenger::Policy::stateless_server(supported, 0));
+  messenger->set_policy(entity_name_t::TYPE_MDS,
+			Messenger::Policy::stateless_server(supported, 0));
 
-  global_print_banner();
+
+  // throttle client traffic
+  Throttle client_throttler(g_ceph_context, "mon_client_bytes",
+			    g_conf->mon_client_bytes);
+  messenger->set_policy_throttler(entity_name_t::TYPE_CLIENT, &client_throttler);
+
+  // throttle daemon traffic
+  // NOTE: actual usage on the leader may multiply by the number of
+  // monitors if they forward large update messages from daemons.
+  Throttle daemon_throttler(g_ceph_context, "mon_daemon_bytes",
+			    g_conf->mon_daemon_bytes);
+  messenger->set_policy_throttler(entity_name_t::TYPE_OSD, &daemon_throttler);
+  messenger->set_policy_throttler(entity_name_t::TYPE_MDS, &daemon_throttler);
 
   cout << "starting " << g_conf->name << " rank " << rank
        << " at " << ipaddr
