@@ -39,6 +39,12 @@ namespace librbd {
     std::string name;
   } snap_info_t;
 
+  typedef struct {
+    std::string client;
+    std::string cookie;
+    std::string address;
+  } locker_t;
+
   typedef rbd_image_info_t image_info_t;
 
   class ProgressContext
@@ -95,19 +101,28 @@ public:
   int parent_info(std::string *parent_poolname, std::string *parent_name,
 		      std::string *parent_snapname);
   int old_format(uint8_t *old);
+  int size(uint64_t *size);
   int features(uint64_t *features);
   int overlap(uint64_t *overlap);
   int copy(IoCtx& dest_io_ctx, const char *destname);
   int copy_with_progress(IoCtx& dest_io_ctx, const char *destname,
 			 ProgressContext &prog_ctx);
 
-  /* cooperative locking */
-  int list_locks(std::set<std::pair<std::string, std::string> > &locks,
-                 bool &exclusive);
+  int flatten();
+  int flatten_with_progress(ProgressContext &prog_ctx);
+  /**
+   * Returns a pair of poolname, imagename for each clone
+   * of this image at the currently set snapshot.
+   */
+  int list_children(std::set<std::pair<std::string, std::string> > *children);
+
+  /* advisory locking (see librbd.h for details) */
+  int list_lockers(std::list<locker_t> *lockers,
+		   bool *exclusive, std::string *tag);
   int lock_exclusive(const std::string& cookie);
-  int lock_shared(const std::string& cookie);
+  int lock_shared(const std::string& cookie, const std::string& tag);
   int unlock(const std::string& cookie);
-  int break_lock(const std::string& other_locker, const std::string& cookie);
+  int break_lock(const std::string& client, const std::string& cookie);
 
   /* snapshots */
   int snap_list(std::vector<snap_info_t>& snaps);
@@ -115,6 +130,9 @@ public:
   int snap_remove(const char *snapname);
   int snap_rollback(const char *snap_name);
   int snap_rollback_with_progress(const char *snap_name, ProgressContext& pctx);
+  int snap_protect(const char *snap_name);
+  int snap_unprotect(const char *snap_name);
+  int snap_is_protected(const char *snap_name, bool *is_protected);
   int snap_set(const char *snap_name);
 
   /* I/O */
