@@ -953,6 +953,11 @@ int librados::IoCtx::aio_write_full(const std::string& oid, librados::AioComplet
   return io_ctx_impl->aio_write_full(obj, c->pc, bl);
 }
 
+int librados::IoCtx::aio_remove(const std::string& oid, librados::AioCompletion *c)
+{
+  return io_ctx_impl->aio_remove(oid, c->pc);
+}
+
 int librados::IoCtx::aio_flush()
 {
   io_ctx_impl->flush_aio_writes();
@@ -1254,6 +1259,11 @@ int librados::Rados::cluster_stat(cluster_stat_t& result)
   return r;
 }
 
+int librados::Rados::cluster_fsid(string *fsid)
+{
+  return client->get_fsid(fsid);
+}
+
 librados::PoolAsyncCompletion *librados::Rados::pool_async_create_completion()
 {
   PoolAsyncCompletionImpl *c = new PoolAsyncCompletionImpl;
@@ -1360,7 +1370,7 @@ extern "C" int rados_conf_read_file(rados_t cluster, const char *path_list)
   librados::RadosClient *client = (librados::RadosClient *)cluster;
   md_config_t *conf = client->cct->_conf;
   std::deque<std::string> parse_errors;
-  int ret = conf->parse_config_files(path_list, &parse_errors, 0);
+  int ret = conf->parse_config_files(path_list, &parse_errors, NULL, 0);
   if (ret)
     return ret;
   conf->parse_env(); // environment variables override
@@ -1447,6 +1457,18 @@ extern "C" int rados_pool_reverse_lookup(rados_t cluster, int64_t id,
   return name.length();
 }
 
+extern "C" int rados_cluster_fsid(rados_t cluster, char *buf,
+				  size_t maxlen)
+{
+  librados::RadosClient *radosp = (librados::RadosClient *)cluster;
+  std::string fsid;
+  radosp->get_fsid(&fsid);
+  if (fsid.length() >= maxlen)
+    return -ERANGE;
+  strcpy(buf, fsid.c_str());
+  return fsid.length();
+}
+
 extern "C" int rados_pool_list(rados_t cluster, char *buf, size_t len)
 {
   librados::RadosClient *client = (librados::RadosClient *)cluster;
@@ -2117,6 +2139,14 @@ extern "C" int rados_aio_write_full(rados_ioctx_t io, const char *o,
   return ctx->aio_write_full(oid, (librados::AioCompletionImpl*)completion, bl);
 }
 
+extern "C" int rados_aio_remove(rados_ioctx_t io, const char *o,
+				rados_completion_t completion)
+{
+  librados::IoCtxImpl *ctx = (librados::IoCtxImpl *)io;
+  object_t oid(o);
+  return ctx->aio_remove(oid, (librados::AioCompletionImpl*)completion);
+}
+
 extern "C" int rados_aio_flush(rados_ioctx_t io)
 {
   librados::IoCtxImpl *ctx = (librados::IoCtxImpl *)io;
