@@ -71,13 +71,12 @@
 
 #include "common/config.h"
 #include "common/errno.h"
-#include "common/DoutStreambuf.h"
 
 #include "perfglue/cpu_profiler.h"
 #include "perfglue/heap_profiler.h"
 
 
-#define DOUT_SUBSYS mds
+#define dout_subsys ceph_subsys_mds
 #undef dout_prefix
 #define dout_prefix *_dout << "mds." << whoami << '.' << incarnation << ' '
 
@@ -88,7 +87,14 @@ MDS::MDS(const std::string &n, Messenger *m, MonClient *mc) :
   Dispatcher(m->cct),
   mds_lock("MDS::mds_lock"),
   timer(m->cct, mds_lock),
-  authorize_handler_registry(new AuthAuthorizeHandlerRegistry(m->cct)),
+  authorize_handler_cluster_registry(new AuthAuthorizeHandlerRegistry(m->cct,
+								      m->cct->_conf->auth_cluster_required.length() ?
+								      m->cct->_conf->auth_cluster_required :
+								      m->cct->_conf->auth_supported)),
+  authorize_handler_service_registry(new AuthAuthorizeHandlerRegistry(m->cct,
+								      m->cct->_conf->auth_service_required.length() ?
+								      m->cct->_conf->auth_service_required :
+								      m->cct->_conf->auth_supported)),
   name(n),
   whoami(-1), incarnation(0),
   standby_for_rank(MDSMap::MDS_NO_STANDBY_PREF),
@@ -96,7 +102,7 @@ MDS::MDS(const std::string &n, Messenger *m, MonClient *mc) :
   standby_replaying(false),
   messenger(m),
   monc(mc),
-  clog(m->cct, messenger, &mc->monmap, mc, LogClient::NO_FLAGS),
+  clog(m->cct, messenger, &mc->monmap, LogClient::NO_FLAGS),
   sessionmap(this) {
 
   orig_argc = 0;
@@ -149,7 +155,8 @@ MDS::MDS(const std::string &n, Messenger *m, MonClient *mc) :
 MDS::~MDS() {
   Mutex::Locker lock(mds_lock);
 
-  delete authorize_handler_registry;
+  delete authorize_handler_service_registry;
+  delete authorize_handler_cluster_registry;
 
   if (mdcache) { delete mdcache; mdcache = NULL; }
   if (mdlog) { delete mdlog; mdlog = NULL; }
@@ -180,7 +187,7 @@ MDS::~MDS() {
   }
   
   if (messenger)
-    messenger->destroy();
+    delete messenger;
 }
 
 void MDS::create_logger()
@@ -440,7 +447,6 @@ int MDS::init(int wanted_state)
   dout(10) << sizeof(xlist<void*>::item) << "\t xlist<>::item   *2=" << 2*sizeof(xlist<void*>::item) << dendl;
 
   messenger->add_dispatcher_tail(this);
-  messenger->add_dispatcher_head(&clog);
 
   // get monmap
   monc->set_messenger(messenger);
@@ -768,13 +774,18 @@ void MDS::handle_command(MMonCommand *m)
     } else dout(0) << "bad migrate_dir syntax" << dendl;
   } 
   else if (m->cmd[0] == "cpu_profiler") {
-    cpu_profiler_handle_command(m->cmd, clog);
+    ostringstream ss;
+    cpu_profiler_handle_command(m->cmd, ss);
+    clog.info() << ss.str();
   }
  else if (m->cmd[0] == "heap") {
    if (!ceph_using_tcmalloc())
      clog.info() << "tcmalloc not enabled, can't use heap profiler commands\n";
-   else
-     ceph_heap_profiler_handle_command(m->cmd, clog);
+   else {
+     ostringstream ss;
+     ceph_heap_profiler_handle_command(m->cmd, ss);
+     clog.info() << ss.str();
+   }
  } else dout(0) << "unrecognized command! " << m->cmd << dendl;
   m->put();
 }
@@ -2011,8 +2022,14 @@ bool MDS::ms_verify_authorizer(Connection *con, int peer_type,
   if (want_state == CEPH_MDS_STATE_DNE)
     return false;
 
-  AuthAuthorizeHandler *authorize_handler =
-    authorize_handler_registry->get_handler(protocol);
+  AuthAuthorizeHandler *authorize_handler = 0;
+  switch (peer_type) {
+  case CEPH_ENTITY_TYPE_MDS:
+    authorize_handler = authorize_handler_cluster_registry->get_handler(protocol);
+    break;
+  default:
+    authorize_handler = authorize_handler_service_registry->get_handler(protocol);
+  }
   if (!authorize_handler) {
     dout(0) << "No AuthAuthorizeHandler found for protocol " << protocol << dendl;
     is_valid = false;
@@ -2034,15 +2051,27 @@ bool MDS::ms_verify_authorizer(Connection *con, int peer_type,
       s = new Session;
       s->inst.addr = con->get_peer_addr();
       s->inst.name = n;
-      dout(10) << " new session " << s << " for " << s->inst << dendl;
+      dout(10) << " new session " << s << " for " << s->inst << " con " << con << dendl;
       con->set_priv(s);
       s->connection = con;
       sessionmap.add_session(s);
     } else {
-      dout(10) << " existing session " << s << " for " << s->inst << dendl;
+      dout(10) << " existing session " << s << " for " << s->inst << " existing con " << s->connection
+	       << ", new/authorizing con " << con << dendl;
       con->set_priv(s->get());
-      s->connection = con;
-   }
+
+      // Wait until we fully accept the connection before setting
+      // s->connection.  In particular, if there are multiple incoming
+      // connection attempts, they will all get their authorizer
+      // validated, but some of them may "lose the race" and get
+      // dropped.  We only want to consider the winner(s).  See
+      // ms_handle_accept().  This is important for Sessions we replay
+      // from the journal on recovery that don't have established
+      // messenger state; we want the con from only the winning
+      // connect attempt(s).  (Normal reconnects that don't follow MDS
+      // recovery are reconnected to the existing con by the
+      // messenger.)
+    }
 
     /*
     s->caps.set_allow_all(caps_info.allow_all);
@@ -2059,3 +2088,15 @@ bool MDS::ms_verify_authorizer(Connection *con, int peer_type,
 };
 
 
+void MDS::ms_handle_accept(Connection *con)
+{
+  Session *s = (Session *)con->get_priv();
+  dout(10) << "ms_handle_accept " << con->get_peer_addr() << " con " << con << " session " << s << dendl;
+  if (s) {
+    s->put();
+    if (s->connection != con) {
+      dout(10) << " session connection " << s->connection << " -> " << con << dendl;
+      s->connection = con;
+    }
+  }
+}
