@@ -12,6 +12,7 @@
  * 
  */
 
+#include <sstream>
 
 #include "AuthMonitor.h"
 #include "Monitor.h"
@@ -29,12 +30,11 @@
 #include "auth/KeyRing.h"
 
 #include "osd/osd_types.h"
-#include "osd/PG.h"  // yuck
 
 #include "common/config.h"
-#include <sstream>
+#include "include/assert.h"
 
-#define DOUT_SUBSYS mon
+#define dout_subsys ceph_subsys_mon
 #undef dout_prefix
 #define dout_prefix _prefix(_dout, mon, paxos->get_version())
 static ostream& _prefix(std::ostream *_dout, Monitor *mon, version_t v) {
@@ -109,13 +109,13 @@ void AuthMonitor::create_initial()
   pending_auth.push_back(inc);
 }
 
-bool AuthMonitor::update_from_paxos()
+void AuthMonitor::update_from_paxos()
 {
   dout(10) << "update_from_paxos()" << dendl;
   version_t paxosv = paxos->get_version();
   version_t keys_ver = mon->key_server.get_ver();
   if (paxosv == keys_ver)
-    return true;
+    return;
   assert(paxosv >= keys_ver);
 
   if (keys_ver != paxos->get_stashed_version()) {
@@ -186,7 +186,10 @@ bool AuthMonitor::update_from_paxos()
   ::encode(mon->key_server, bl);
   paxos->stash_latest(paxosv, bl);
 
-  return true;
+  unsigned max = g_conf->paxos_max_join_drift * 2;
+  if (mon->is_leader() &&
+      paxosv > max)
+    paxos->trim_to(paxosv - max);
 }
 
 void AuthMonitor::increase_max_global_id()
@@ -220,7 +223,7 @@ void AuthMonitor::encode_pending(bufferlist &bl)
   for (vector<Incremental>::iterator p = pending_auth.begin();
        p != pending_auth.end();
        p++)
-    p->encode(bl);
+    p->encode(bl, mon->get_quorum_features());
 }
 
 bool AuthMonitor::preprocess_query(PaxosServiceMessage *m)
@@ -342,7 +345,14 @@ bool AuthMonitor::prep_auth(MAuth *m, bool paxos_writable)
       goto reply;
     }
 
-    int type = mon->auth_supported.pick(supported);
+    int type;
+
+    if (entity_name.get_type() == CEPH_ENTITY_TYPE_MON ||
+	entity_name.get_type() == CEPH_ENTITY_TYPE_OSD ||
+	entity_name.get_type() == CEPH_ENTITY_TYPE_MDS)
+      type = mon->auth_cluster_required.pick(supported);
+    else
+      type = mon->auth_service_required.pick(supported);
     s->auth_handler = get_auth_service_handler(type, g_ceph_context, &mon->key_server);
     if (!s->auth_handler) {
       ret = -ENOTSUP;
@@ -395,7 +405,7 @@ bool AuthMonitor::prep_auth(MAuth *m, bool paxos_writable)
     }
     if (caps_info.caps.length()) {
       bufferlist::iterator iter = caps_info.caps.begin();
-      s->caps.parse(iter);
+      s->caps.parse(iter, NULL);
       s->caps.set_auid(auid);
     }
   } catch (const buffer::error &err) {
@@ -428,11 +438,21 @@ bool AuthMonitor::preprocess_command(MMonCommand *m)
   if (m->cmd.size() > 1) {
     if (m->cmd[1] == "add" ||
         m->cmd[1] == "del" ||
-	m->cmd[1] == "caps" ||
-        m->cmd[1] == "list") {
+	m->cmd[1] == "get-or-create" ||
+	m->cmd[1] == "get-or-create-key" ||
+	m->cmd[1] == "caps") {
       return false;
     }
-    else if (m->cmd[1] == "export") {
+
+    MonSession *session = m->get_session();
+    if (!session ||
+	(!session->caps.get_allow_all() &&
+	 !mon->_allowed_command(session, m->cmd))) {
+      mon->reply_command(m, -EACCES, "access denied", rdata, paxos->get_version());
+      return true;
+    }
+
+    if (m->cmd[1] == "export") {
       KeyRing keyring;
       export_keyring(keyring);
       if (m->cmd.size() > 2) {
@@ -442,7 +462,7 @@ bool AuthMonitor::preprocess_command(MMonCommand *m)
 	  if (keyring.get_auth(ename, eauth)) {
 	    KeyRing kr;
 	    kr.add(ename, eauth);
-	    ::encode(kr, rdata);
+	    kr.encode_plaintext(rdata);
 	    ss << "export " << eauth;
 	    r = 0;
 	  } else {
@@ -454,7 +474,7 @@ bool AuthMonitor::preprocess_command(MMonCommand *m)
 	  r = -EINVAL;
 	}
       } else {
-	::encode(keyring, rdata);
+	keyring.encode_plaintext(rdata);
 	ss << "exported master keyring";
 	r = 0;
       }
@@ -472,13 +492,13 @@ bool AuthMonitor::preprocess_command(MMonCommand *m)
 	  r = -ENOENT;
 	} else {
 	  keyring.add(entity, entity_auth);
-	  ::encode(keyring, rdata);
+	  keyring.encode_plaintext(rdata);
 	  ss << "exported keyring for " << m->cmd[2];
 	  r = 0;
 	}
       }
     }
-    else if ((m->cmd[1] == "print-key" || m->cmd[1] == "print_key") &&
+    else if ((m->cmd[1] == "print-key" || m->cmd[1] == "print_key" || m->cmd[1] == "get-key") &&
 	     m->cmd.size() == 3) {
       EntityName ename;
       if (!ename.from_str(m->cmd[2])) {
@@ -495,6 +515,11 @@ bool AuthMonitor::preprocess_command(MMonCommand *m)
       ss << auth.key;
       r = 0;      
     }
+    else if (m->cmd[1] == "list") {
+      mon->key_server.list_secrets(ss);
+      r = 0;
+      goto done;
+    }
     else {
       auth_usage(ss);
       r = -EINVAL;
@@ -533,9 +558,18 @@ void AuthMonitor::import_keyring(KeyRing& keyring)
 bool AuthMonitor::prepare_command(MMonCommand *m)
 {
   stringstream ss;
+  bufferlist rdata;
   string rs;
   int err = -EINVAL;
 
+  MonSession *session = m->get_session();
+  if (!session ||
+      (!session->caps.get_allow_all() &&
+       !mon->_allowed_command(session, m->cmd))) {
+    mon->reply_command(m, -EACCES, "access denied", rdata, paxos->get_version());
+    return true;
+  }
+
   // nothing here yet
   if (m->cmd.size() > 1) {
     if (m->cmd[1] == "import") {
@@ -602,6 +636,81 @@ bool AuthMonitor::prepare_command(MMonCommand *m)
       paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, paxos->get_version()));
       return true;
     }
+    else if ((m->cmd[1] == "get-or-create-key" ||
+	      m->cmd[1] == "get-or-create") &&
+	     m->cmd.size() >= 3) {
+      // auth get-or-create <name> [mon osdcapa osd osdcapb ...]
+      EntityName entity;
+      if (!entity.from_str(m->cmd[2])) {
+	ss << "bad entity name";
+	err = -EINVAL;
+	goto done;
+      }
+
+      // do we have it?
+      EntityAuth entity_auth;
+      if (mon->key_server.get_auth(entity, entity_auth)) {
+	for (unsigned i=3; i + 1<m->cmd.size(); i += 2) {
+	  string sys = m->cmd[i];
+	  bufferlist cap;
+	  ::encode(m->cmd[i+1], cap);
+	  if (entity_auth.caps.count(sys) == 0 ||
+	      !entity_auth.caps[sys].contents_equal(cap)) {
+	    ss << "key for " << entity << " exists but cap " << sys << " does not match";
+	    err = -EINVAL;
+	    goto done;
+	  }
+	}
+
+	if (m->cmd[1] == "get-or-create-key") {
+	  ss << entity_auth.key;
+	} else {
+	  KeyRing kr;
+	  kr.add(entity, entity_auth.key);
+	  kr.encode_plaintext(rdata);
+	}
+	err = 0;
+	goto done;
+      }
+
+      // ...or are we about to?
+      for (vector<Incremental>::iterator p = pending_auth.begin();
+	   p != pending_auth.end();
+	   ++p) {
+	if (p->inc_type == AUTH_DATA) {
+	  KeyServerData::Incremental auth_inc;
+	  bufferlist::iterator q = p->auth_data.begin();
+	  ::decode(auth_inc, q);
+	  if (auth_inc.op == KeyServerData::AUTH_INC_ADD &&
+	      auth_inc.name == entity) {
+	    paxos->wait_for_commit(new C_RetryMessage(this, m));
+	    return true;
+	  }
+	}
+      }
+
+      // create it
+      KeyServerData::Incremental auth_inc;
+      auth_inc.op = KeyServerData::AUTH_INC_ADD;
+      auth_inc.name = entity;
+      auth_inc.auth.key.create(g_ceph_context, CEPH_CRYPTO_AES);
+      for (unsigned i=3; i + 1<m->cmd.size(); i += 2)
+	::encode(m->cmd[i+1], auth_inc.auth.caps[m->cmd[i]]);
+
+      push_cephx_inc(auth_inc);
+
+      if (m->cmd[1] == "get-or-create-key") {
+	ss << auth_inc.auth.key;
+      } else {
+	KeyRing kr;
+	kr.add(entity, auth_inc.auth.key);
+	kr.encode_plaintext(rdata);
+      }
+
+      getline(ss, rs);
+      paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, rdata, paxos->get_version()));
+      return true;
+    }
     else if (m->cmd[1] == "caps" && m->cmd.size() >= 3) {
       KeyServerData::Incremental auth_inc;
       if (!auth_inc.name.from_str(m->cmd[2])) {
@@ -646,11 +755,6 @@ bool AuthMonitor::prepare_command(MMonCommand *m)
       paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, paxos->get_version()));
       return true;
     }
-    else if (m->cmd[1] == "list") {
-      mon->key_server.list_secrets(ss);
-      err = 0;
-      goto done;
-    }
     else {
       auth_usage(ss);
     }
@@ -660,7 +764,7 @@ bool AuthMonitor::prepare_command(MMonCommand *m)
 
 done:
   getline(ss, rs, '\0');
-  mon->reply_command(m, err, rs, paxos->get_version());
+  mon->reply_command(m, err, rs, rdata, paxos->get_version());
   return false;
 }
 
