@@ -34,6 +34,7 @@
 #include <sstream>
 
 #include "common/config.h"
+#include "include/assert.h"
 
 #define dout_subsys ceph_subsys_mon
 #undef dout_prefix
@@ -489,6 +490,9 @@ void MDSMonitor::on_active()
 {
   tick();
   update_logger();
+
+  if (mon->is_leader())
+    mon->clog.info() << "mdsmap " << mdsmap << "\n";
 }
 
 void MDSMonitor::get_health(list<pair<health_status_t, string> >& summary,
@@ -497,12 +501,28 @@ void MDSMonitor::get_health(list<pair<health_status_t, string> >& summary,
   mdsmap.get_health(summary, detail);
 }
 
+void MDSMonitor::dump_info(Formatter *f)
+{
+  f->open_object_section("mdsmap");
+  mdsmap.dump(f);
+  f->close_section();
+}
+
 bool MDSMonitor::preprocess_command(MMonCommand *m)
 {
   int r = -1;
   bufferlist rdata;
   stringstream ss;
 
+  MonSession *session = m->get_session();
+  if (!session ||
+      (!session->caps.get_allow_all() &&
+       !session->caps.check_privileges(PAXOS_MDSMAP, MON_CAP_R) &&
+       !mon->_allowed_command(session, m->cmd))) {
+    mon->reply_command(m, -EACCES, "access denied", rdata, paxos->get_version());
+    return true;
+  }
+
   vector<const char*> args;
   for (unsigned i = 1; i < m->cmd.size(); i++)
     args.push_back(m->cmd[i].c_str());
@@ -716,6 +736,15 @@ bool MDSMonitor::prepare_command(MMonCommand *m)
   stringstream ss;
   bufferlist rdata;
 
+  MonSession *session = m->get_session();
+  if (!session ||
+      (!session->caps.get_allow_all() &&
+       !session->caps.check_privileges(PAXOS_MDSMAP, MON_CAP_W) &&
+       !mon->_allowed_command(session, m->cmd))) {
+    mon->reply_command(m, -EACCES, "access denied", rdata, paxos->get_version());
+    return true;
+  }
+
   if (m->cmd.size() > 1) {
     if ((m->cmd[1] == "stop" || m->cmd[1] == "deactivate") && m->cmd.size() > 2) {
       int who = atoi(m->cmd[2].c_str());
@@ -862,18 +891,23 @@ bool MDSMonitor::prepare_command(MMonCommand *m)
       r = pending_mdsmap.remove_data_pg_pool(poolid);
       if (r == 0)
 	ss << "removed data pool " << poolid << " from mdsmap";
-    } else if (m->cmd[1] == "newfs" && m->cmd.size() == 4) {
+    } else if (m->cmd[1] == "newfs" && m->cmd.size() >= 4) {
       MDSMap newmap;
       int metadata = atoi(m->cmd[2].c_str());
       int data = atoi(m->cmd[3].c_str());
-      pending_mdsmap = newmap;
-      pending_mdsmap.epoch = mdsmap.epoch + 1;
-      create_new_fs(pending_mdsmap, metadata, data);
-      ss << "new fs with metadata pool " << metadata << " and data pool " << data;
-      string rs;
-      getline(ss, rs);
-      paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, paxos->get_version()));
-      return true;
+      if (m->cmd.size() < 5 || m->cmd[4] != "--yes-i-really-mean-it") {
+	ss << "this is DANGEROUS and will wipe out the mdsmap's fs, and may clobber data in the new pools you specify.  add --yes-i-really-mean-it if you do.";
+	r = -EPERM;
+      } else {
+	pending_mdsmap = newmap;
+	pending_mdsmap.epoch = mdsmap.epoch + 1;
+	create_new_fs(pending_mdsmap, metadata, data);
+	ss << "new fs with metadata pool " << metadata << " and data pool " << data;
+	string rs;
+	getline(ss, rs);
+	paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, paxos->get_version()));
+	return true;
+      }
     }    
   }
   if (r == -EINVAL) 
