@@ -18,6 +18,7 @@
 #include <signal.h>
 
 #include "Monitor.h"
+#include "common/version.h"
 
 #include "osd/OSDMap.h"
 
@@ -65,7 +66,7 @@
 
 #include "osd/OSDMap.h"
 
-#include "auth/AuthSupported.h"
+#include "auth/AuthMethodList.h"
 #include "auth/KeyRing.h"
 
 #include "common/config.h"
@@ -101,7 +102,12 @@ Monitor::Monitor(CephContext* cct_, string nm, MonitorStore *s, Messenger *m, Mo
   monmap(map),
   clog(cct_, messenger, monmap, LogClient::FLAG_MON),
   key_server(cct, &keyring),
-  auth_supported(cct),
+  auth_cluster_required(cct,
+			cct->_conf->auth_cluster_required.length() ?
+			cct->_conf->auth_cluster_required : cct->_conf->auth_supported),
+  auth_service_required(cct,
+		       cct->_conf->auth_service_required.length() ?
+		       cct->_conf->auth_service_required : cct->_conf->auth_supported),
   store(s),
   
   state(STATE_PROBING),
@@ -270,6 +276,15 @@ int Monitor::init()
     cluster_logger = pcb.create_perf_counters();
   }
 
+  // verify cluster_uuid
+  {
+    int r = check_fsid();
+    if (r == -ENOENT)
+      r = write_fsid();
+    if (r < 0)
+      return r;
+  }
+
   // open compatset
   {
     bufferlist bl;
@@ -924,6 +939,11 @@ void Monitor::start_election()
 void Monitor::win_standalone_election()
 {
   dout(1) << "win_standalone_election" << dendl;
+
+  // bump election epoch, in case the previous epoch included other
+  // monitors; we need to be able to make the distinction.
+  elector.advance_epoch();
+
   rank = monmap->get_rank(name);
   assert(rank == 0);
   set<int> q;
@@ -1086,10 +1106,14 @@ void Monitor::_mon_status(ostream& ss)
   jf.flush(ss);
 }
 
-void Monitor::get_health(string& status, bufferlist *detailbl)
+void Monitor::get_health(string& status, bufferlist *detailbl, Formatter *f)
 {
   list<pair<health_status_t,string> > summary;
   list<pair<health_status_t,string> > detail;
+
+  if (f)
+    f->open_object_section("health");
+
   for (vector<PaxosService*>::iterator p = paxos_service.begin();
        p != paxos_service.end();
        p++) {
@@ -1097,9 +1121,17 @@ void Monitor::get_health(string& status, bufferlist *detailbl)
     s->get_health(summary, detailbl ? &detail : NULL);
   }
 
+  if (f)
+    f->open_array_section("summary");
   stringstream ss;
   health_status_t overall = HEALTH_OK;
   if (!summary.empty()) {
+    if (f) {
+      f->open_object_section("item");
+      f->dump_stream("severity") <<  summary.front().first;
+      f->dump_string("summary", summary.front().second);
+      f->close_section();
+    }
     ss << ' ';
     while (!summary.empty()) {
       if (overall > summary.front().first)
@@ -1110,15 +1142,28 @@ void Monitor::get_health(string& status, bufferlist *detailbl)
 	ss << "; ";
     }
   }
+  if (f)
+    f->close_section();
   stringstream fss;
   fss << overall;
   status = fss.str() + ss.str();
+  if (f)
+    f->dump_stream("overall_status") << overall;
 
+  if (f)
+    f->open_array_section("detail");
   while (!detail.empty()) {
+    if (f)
+      f->dump_string("item", detail.front().second);
     detailbl->append(detail.front().second);
     detailbl->append('\n');
     detail.pop_front();
   }
+  if (f)
+    f->close_section();
+
+  if (f)
+    f->close_section();
 }
 
 void Monitor::handle_command(MMonCommand *m)
@@ -1232,7 +1277,7 @@ void Monitor::handle_command(MMonCommand *m)
       }
       // reply with the status for all the components
       string health;
-      get_health(health, NULL);
+      get_health(health, NULL, NULL);
       stringstream ss;
       ss << "   health " << health << "\n";
       ss << "   monmap " << *monmap << ", election epoch " << get_epoch() << ", quorum " << get_quorum()
@@ -1243,6 +1288,49 @@ void Monitor::handle_command(MMonCommand *m)
       rs = ss.str();
       r = 0;
     }
+    if (m->cmd[0] == "report") {
+      if (!access_r) {
+	r = -EACCES;
+	rs = "access denied";
+	goto out;
+      }
+
+      JSONFormatter jf(true);
+
+      jf.open_object_section("report");
+      jf.dump_string("version", ceph_version_to_str());
+      jf.dump_string("commit", git_version_to_str());
+      jf.dump_stream("timestamp") << ceph_clock_now(NULL);
+
+      string d;
+      for (unsigned i = 1; i < m->cmd.size(); i++) {
+	if (i > 1)
+	  d += " ";
+	d += m->cmd[i];
+      }
+      jf.dump_string("tag", d);
+
+      string hs;
+      get_health(hs, NULL, &jf);
+      
+      monmon()->dump_info(&jf);
+      osdmon()->dump_info(&jf);
+      mdsmon()->dump_info(&jf);
+      pgmon()->dump_info(&jf);
+
+      jf.close_section();
+      stringstream ss;
+      jf.flush(ss);
+
+      bufferlist bl;
+      bl.append("-------- BEGIN REPORT --------\n");
+      bl.append(ss);
+      ostringstream ss2;
+      ss2 << "\n-------- END REPORT " << bl.crc32c(6789) << " --------\n";
+      rdata.append(bl);
+      rdata.append(ss2.str());
+      r = 0;
+    }
     if (m->cmd[0] == "quorum_status") {
       if (!access_r) {
 	r = -EACCES;
@@ -1277,7 +1365,7 @@ void Monitor::handle_command(MMonCommand *m)
 	rs = "access denied";
 	goto out;
       }
-      get_health(rs, (m->cmd.size() > 1) ? &rdata : NULL);
+      get_health(rs, (m->cmd.size() > 1) ? &rdata : NULL, NULL);
       r = 0;
     }
     if (m->cmd[0] == "heap") {
@@ -1288,8 +1376,11 @@ void Monitor::handle_command(MMonCommand *m)
       }
       if (!ceph_using_tcmalloc())
 	rs = "tcmalloc not enabled, can't use heap profiler commands\n";
-      else
-	ceph_heap_profiler_handle_command(m->cmd, clog);
+      else {
+	ostringstream ss;
+	ceph_heap_profiler_handle_command(m->cmd, ss);
+	rs = ss.str();
+      }
     }
     if (m->cmd[0] == "quorum") {
       if (!access_all) {
@@ -2007,6 +2098,45 @@ void Monitor::tick()
   new_tick();
 }
 
+int Monitor::check_fsid()
+{
+  ostringstream ss;
+  ss << monmap->get_fsid();
+  string us = ss.str();
+  bufferlist ebl;
+  int r = store->get_bl_ss(ebl, "cluster_uuid", 0);
+  if (r < 0)
+    return r;
+
+  string es(ebl.c_str(), ebl.length());
+
+  // only keep the first line
+  size_t pos = es.find_first_of('\n');
+  if (pos != string::npos)
+    es.resize(pos);
+
+  dout(10) << "check_fsid cluster_uuid contains '" << es << "'" << dendl;
+  if (es.length() < us.length() ||
+      strncmp(us.c_str(), es.c_str(), us.length()) != 0) {
+    derr << "error: cluster_uuid file exists with value '" << es
+	 << "', != our uuid " << monmap->get_fsid() << dendl;
+    return -EEXIST;
+  }
+
+  return 0;
+}
+
+int Monitor::write_fsid()
+{
+  ostringstream ss;
+  ss << monmap->get_fsid() << "\n";
+  string us = ss.str();
+
+  bufferlist b;
+  b.append(us);
+  return store->put_bl_ss(b, "cluster_uuid", 0);
+}
+
 /*
  * this is the closest thing to a traditional 'mkfs' for ceph.
  * initialize the monitor state machines to their initial values.
@@ -2020,10 +2150,15 @@ int Monitor::mkfs(bufferlist& osdmapbl)
     return err;
   }
 
+  // verify cluster fsid
+  int r = check_fsid();
+  if (r < 0 && r != -ENOENT)
+    return r;
+
   bufferlist magicbl;
   magicbl.append(CEPH_MON_ONDISK_MAGIC);
   magicbl.append("\n");
-  int r = store->put_bl_ss(magicbl, "magic", 0);
+  r = store->put_bl_ss(magicbl, "magic", 0);
   if (r < 0)
     return r;
 
@@ -2065,6 +2200,12 @@ int Monitor::mkfs(bufferlist& osdmapbl)
   keyring.encode_plaintext(keyringbl);
   store->put_bl_ss(keyringbl, "mkfs", "keyring");
 
+  // sync and write out fsid to indicate completion.
+  store->sync();
+  r = write_fsid();
+  if (r < 0)
+    return r;
+
   return 0;
 }
 
@@ -2095,7 +2236,7 @@ bool Monitor::ms_get_authorizer(int service_id, AuthAuthorizer **authorizer, boo
   if (service_id != CEPH_ENTITY_TYPE_MON)
     return false;
 
-  if (!auth_supported.is_supported_auth(CEPH_AUTH_CEPHX))
+  if (!auth_cluster_required.is_supported_auth(CEPH_AUTH_CEPHX))
     return false;
 
   CephXServiceTicketInfo auth_ticket_info;
@@ -2156,7 +2297,7 @@ bool Monitor::ms_verify_authorizer(Connection *con, int peer_type,
     return false;
 
   if (peer_type == CEPH_ENTITY_TYPE_MON &&
-      auth_supported.is_supported_auth(CEPH_AUTH_CEPHX)) {
+      auth_cluster_required.is_supported_auth(CEPH_AUTH_CEPHX)) {
     // monitor, and cephx is enabled
     isvalid = false;
     if (protocol == CEPH_AUTH_CEPHX) {
