@@ -38,7 +38,8 @@
 #include "common/LogClient.h"
 
 #include "auth/cephx/CephxKeyServer.h"
-#include "auth/AuthSupported.h"
+#include "auth/AuthMethodList.h"
+#include "auth/KeyRing.h"
 
 #include "perfglue/heap_profiler.h"
 
@@ -85,7 +86,6 @@ class AdminSocketHook;
 class MMonGetMap;
 class MMonGetVersion;
 class MMonProbe;
-class MMonObserve;
 class MMonSubscribe;
 class MAuthRotating;
 class MRoute;
@@ -102,6 +102,11 @@ public:
   Mutex lock;
   SafeTimer timer;
   
+  /// true if we have ever joined a quorum.  if false, we are either a
+  /// new cluster, a newly joining monitor, or a just-upgraded
+  /// monitor.
+  bool has_ever_joined;
+
   PerfCounters *logger, *cluster_logger;
   bool cluster_logger_registered;
 
@@ -110,10 +115,14 @@ public:
 
   MonMap *monmap;
 
+  set<entity_addr_t> extra_probe_peers;
+
   LogClient clog;
+  KeyRing keyring;
   KeyServer key_server;
 
-  AuthSupported auth_supported;
+  AuthMethodList auth_cluster_required;
+  AuthMethodList auth_service_required;
 
   CompatSet features;
 
@@ -169,6 +178,7 @@ private:
   set<int> quorum;       // current active set of monitors (if !starting)
   utime_t leader_since;  // when this monitor became the leader, if it is the leader
   utime_t exited_quorum; // time detected as not in quorum; 0 if in
+  unsigned quorum_features;  ///< intersection of quorum member feature bits
 
   set<string> outside_quorum;
   entity_inst_t slurp_source;
@@ -198,12 +208,22 @@ public:
   epoch_t get_epoch();
   int get_leader() { return leader; }
   const set<int>& get_quorum() { return quorum; }
+  set<string> get_quorum_names() {
+    set<string> q;
+    for (set<int>::iterator p = quorum.begin(); p != quorum.end(); ++p)
+      q.insert(monmap->get_name(*p));
+    return q;
+  }
+  unsigned get_quorum_features() const {
+    return quorum_features;
+  }
 
   void bootstrap();
   void reset();
   void start_election();
   void win_standalone_election();
-  void win_election(epoch_t epoch, set<int>& q);         // end election (called by Elector)
+  void win_election(epoch_t epoch, set<int>& q,
+		    unsigned features);         // end election (called by Elector)
   void lose_election(epoch_t epoch, set<int>& q, int l); // end election (called by Elector)
   void finish_election();
 
@@ -217,11 +237,29 @@ public:
   Paxos *get_paxos_by_name(const string& name);
   PaxosService *get_paxos_service_by_name(const string& name);
 
-  class PGMonitor *pgmon() { return (class PGMonitor *)paxos_service[PAXOS_PGMAP]; }
-  class MDSMonitor *mdsmon() { return (class MDSMonitor *)paxos_service[PAXOS_MDSMAP]; }
-  class MonmapMonitor *monmon() { return (class MonmapMonitor *)paxos_service[PAXOS_MONMAP]; }
-  class OSDMonitor *osdmon() { return (class OSDMonitor *)paxos_service[PAXOS_OSDMAP]; }
-  class AuthMonitor *authmon() { return (class AuthMonitor *)paxos_service[PAXOS_AUTH]; }
+  class PGMonitor *pgmon() {
+    return (class PGMonitor *)paxos_service[PAXOS_PGMAP];
+  }
+
+  class MDSMonitor *mdsmon() {
+    return (class MDSMonitor *)paxos_service[PAXOS_MDSMAP];
+  }
+
+  class MonmapMonitor *monmon() {
+    return (class MonmapMonitor *)paxos_service[PAXOS_MONMAP];
+  }
+
+  class OSDMonitor *osdmon() {
+    return (class OSDMonitor *)paxos_service[PAXOS_OSDMAP];
+  }
+
+  class AuthMonitor *authmon() {
+    return (class AuthMonitor *)paxos_service[PAXOS_AUTH];
+  }
+
+  class LogMonitor *logmon() {
+    return (class LogMonitor*) paxos_service[PAXOS_LOG];
+  }
 
   friend class Paxos;
   friend class OSDMonitor;
@@ -247,20 +285,43 @@ public:
   bool _allowed_command(MonSession *s, const vector<std::string>& cmd);
   void _mon_status(ostream& ss);
   void _quorum_status(ostream& ss);
+  void _add_bootstrap_peer_hint(string cmd, string args, ostream& ss);
   void handle_command(class MMonCommand *m);
-  void handle_observe(MMonObserve *m);
   void handle_route(MRoute *m);
 
+  /**
+   * Generate health report
+   *
+   * @param status one-line status summary
+   * @param detailbl optional bufferlist* to fill with a detailed report
+   */
+  void get_health(string& status, bufferlist *detailbl, Formatter *f);
+
   void reply_command(MMonCommand *m, int rc, const string &rs, version_t version);
   void reply_command(MMonCommand *m, int rc, const string &rs, bufferlist& rdata, version_t version);
 
   void handle_probe(MMonProbe *m);
+  /**
+   * Handle a Probe Operation, replying with our name, quorum and known versions.
+   *
+   * We use the MMonProbe message class for anything and everything related with
+   * Monitor probing. One of the operations relates directly with the probing
+   * itself, in which we receive a probe request and to which we reply with
+   * our name, our quorum and the known versions for each Paxos service. Thus the
+   * redundant function name. This reply will obviously be sent to the one
+   * probing/requesting these infos.
+   *
+   * @todo Add @pre and @post
+   *
+   * @param m A Probe message, with an operation of type Probe.
+   */
   void handle_probe_probe(MMonProbe *m);
   void handle_probe_reply(MMonProbe *m);
   void handle_probe_slurp(MMonProbe *m);
   void handle_probe_slurp_latest(MMonProbe *m);
   void handle_probe_data(MMonProbe *m);
-  /* Given an MMonProbe and associated Paxos machine, create a reply,
+  /**
+   * Given an MMonProbe and associated Paxos machine, create a reply,
    * fill it with the missing Paxos states and current commit pointers
    *
    * @param m The incoming MMonProbe. We use this to determine the range
@@ -302,11 +363,14 @@ public:
     MMonCommand *m;
     int rc;
     string rs;
+    bufferlist rdata;
     version_t version;
     C_Command(Monitor *_mm, MMonCommand *_m, int r, string s, version_t v) :
       mon(_mm), m(_m), rc(r), rs(s), version(v){}
+    C_Command(Monitor *_mm, MMonCommand *_m, int r, string s, bufferlist rd, version_t v) :
+      mon(_mm), m(_m), rc(r), rs(s), rdata(rd), version(v){}
     void finish(int r) {
-      mon->reply_command(m, rc, rs, version);
+      mon->reply_command(m, rc, rs, rdata, version);
     }
   };
 
@@ -339,11 +403,13 @@ public:
   bool ms_handle_reset(Connection *con);
   void ms_handle_remote_reset(Connection *con) {}
 
+  void extract_save_mon_key(KeyRing& keyring);
+
  public:
   Monitor(CephContext *cct_, string nm, MonitorStore *s, Messenger *m, MonMap *map);
   ~Monitor();
 
-  void init();
+  int init();
   void shutdown();
   void tick();
 
@@ -353,7 +419,21 @@ public:
 
   int mkfs(bufferlist& osdmapbl);
 
-  void do_admin_command(std::string command, ostream& ss);
+  /**
+   * check cluster_fsid file
+   *
+   * @return EEXIST if file exists and doesn't match, 0 on match, or negative error code
+   */
+  int check_fsid();
+
+  /**
+   * write cluster_fsid file
+   *
+   * @return 0 on success, or negative error code
+   */
+  int write_fsid();
+
+  void do_admin_command(std::string command, std::string args, ostream& ss);
 
 private:
   // don't allow copying
