@@ -859,6 +859,12 @@ bool PGMonitor::check_down_pgs()
   return ret;
 }
 
+void PGMonitor::dump_info(Formatter *f)
+{
+  f->open_object_section("pgmap");
+  pg_map.dump(f);
+  f->close_section();
+}
 
 bool PGMonitor::preprocess_command(MMonCommand *m)
 {
@@ -1151,14 +1157,40 @@ bool PGMonitor::prepare_command(MMonCommand *m)
   return false;
 }
 
-static void note_stuck_detail(hash_map<pg_t,pg_stat_t>& stuck_pgs,
+static void note_stuck_detail(enum PGMap::StuckPG what,
+			      hash_map<pg_t,pg_stat_t>& stuck_pgs,
 			      list<pair<health_status_t,string> > *detail)
 {
   for (hash_map<pg_t,pg_stat_t>::iterator p = stuck_pgs.begin();
        p != stuck_pgs.end();
        ++p) {
     ostringstream ss;
-    ss << "pg " << p->first << " is stuck " << pg_state_string(p->second.state)
+    utime_t since;
+    const char *whatname = 0;
+    switch (what) {
+    case PGMap::STUCK_INACTIVE:
+      since = p->second.last_active;
+      whatname = "inactive";
+      break;
+    case PGMap::STUCK_UNCLEAN:
+      since = p->second.last_clean;
+      whatname = "unclean";
+      break;
+    case PGMap::STUCK_STALE:
+      since = p->second.last_unstale;
+      whatname = "stale";
+      break;
+    default:
+      assert(0);
+    }
+    ss << "pg " << p->first << " is stuck " << whatname;
+    if (since == utime_t()) {
+      ss << " since forever";
+    }else {
+      utime_t dur = ceph_clock_now(g_ceph_context) - since;
+      ss << " for " << dur;
+    }
+    ss << ", current state " << pg_state_string(p->second.state)
        << ", last acting " << p->second.acting;
     detail->push_back(make_pair(HEALTH_WARN, ss.str()));
   }
@@ -1201,7 +1233,7 @@ void PGMonitor::get_health(list<pair<health_status_t,string> >& summary,
   if (!stuck_pgs.empty()) {
     note["stuck inactive"] = stuck_pgs.size();
     if (detail)
-      note_stuck_detail(stuck_pgs, detail);
+      note_stuck_detail(PGMap::STUCK_INACTIVE, stuck_pgs, detail);
   }
   stuck_pgs.clear();
 
@@ -1209,7 +1241,7 @@ void PGMonitor::get_health(list<pair<health_status_t,string> >& summary,
   if (!stuck_pgs.empty()) {
     note["stuck unclean"] = stuck_pgs.size();
     if (detail)
-      note_stuck_detail(stuck_pgs, detail);
+      note_stuck_detail(PGMap::STUCK_UNCLEAN, stuck_pgs, detail);
   }
   stuck_pgs.clear();
 
@@ -1217,7 +1249,7 @@ void PGMonitor::get_health(list<pair<health_status_t,string> >& summary,
   if (!stuck_pgs.empty()) {
     note["stuck stale"] = stuck_pgs.size();
     if (detail)
-      note_stuck_detail(stuck_pgs, detail);
+      note_stuck_detail(PGMap::STUCK_STALE, stuck_pgs, detail);
   }
 
   if (!note.empty()) {
