@@ -17,16 +17,17 @@
 #include "MonitorStore.h"
 
 #include "messages/MMonPaxos.h"
-#include "messages/MMonObserveNotify.h"
 
 #include "common/config.h"
+#include "include/assert.h"
 
 #define dout_subsys ceph_subsys_paxos
 #undef dout_prefix
 #define dout_prefix _prefix(_dout, mon, mon->name, mon->rank, machine_name, state, first_committed, last_committed)
-static ostream& _prefix(std::ostream *_dout, Monitor *mon, const string& name, int rank,
-			const char *machine_name, int state,
-			version_t first_committed, version_t last_committed) {
+static ostream& _prefix(std::ostream *_dout, Monitor *mon, const string& name,
+		        int rank, const char *machine_name, int state,
+			version_t first_committed, version_t last_committed)
+{
   return *_dout << "mon." << name << "@" << rank
 		<< "(" << mon->get_state_name() << ")"
 		<< ".paxos(" << machine_name << " " << Paxos::get_statename(state)
@@ -90,8 +91,8 @@ void Paxos::collect(version_t oldpn)
        ++p) {
     if (*p == mon->rank) continue;
     
-    MMonPaxos *collect = new MMonPaxos(mon->get_epoch(), MMonPaxos::OP_COLLECT, machine_id,
-				       ceph_clock_now(g_ceph_context));
+    MMonPaxos *collect = new MMonPaxos(mon->get_epoch(), MMonPaxos::OP_COLLECT,
+				       machine_id, ceph_clock_now(g_ceph_context));
     collect->last_committed = last_committed;
     collect->first_committed = first_committed;
     collect->pn = accepted_pn;
@@ -125,13 +126,14 @@ void Paxos::handle_collect(MMonPaxos *collect)
     // ok, accept it
     accepted_pn = collect->pn;
     accepted_pn_from = collect->pn_from;
-    dout(10) << "accepting pn " << accepted_pn << " from " << accepted_pn_from << dendl;
+    dout(10) << "accepting pn " << accepted_pn << " from " 
+	     << accepted_pn_from << dendl;
     mon->store->put_int(accepted_pn, machine_name, "accepted_pn");
   } else {
     // don't accept!
-    dout(10) << "NOT accepting pn " << collect->pn << " from " << collect->pn_from 
-	     << ", we already accepted " << accepted_pn << " from " << accepted_pn_from 
-	     << dendl;
+    dout(10) << "NOT accepting pn " << collect->pn << " from " << collect->pn_from
+	     << ", we already accepted " << accepted_pn
+	     << " from " << accepted_pn_from << dendl;
   }
   last->pn = accepted_pn;
   last->pn_from = accepted_pn_from;
@@ -146,8 +148,8 @@ void Paxos::handle_collect(MMonPaxos *collect)
   if (mon->store->exists_bl_sn(machine_name, last_committed+1)) {
     mon->store->get_bl_sn(bl, machine_name, last_committed+1);
     assert(bl.length() > 0);
-    dout(10) << " sharing our accepted but uncommitted value for " << last_committed+1 
-	     << " (" << bl.length() << " bytes)" << dendl;
+    dout(10) << " sharing our accepted but uncommitted value for " 
+	     << last_committed+1 << " (" << bl.length() << " bytes)" << dendl;
     last->values[last_committed+1] = bl;
     last->uncommitted_pn = accepted_pn;
   }
@@ -157,11 +159,13 @@ void Paxos::handle_collect(MMonPaxos *collect)
   collect->put();
 }
 
-void Paxos::share_state(MMonPaxos *m, version_t peer_first_committed, version_t peer_last_committed)
+void Paxos::share_state(MMonPaxos *m, version_t peer_first_committed,
+			version_t peer_last_committed)
 {
   assert(peer_last_committed < last_committed);
 
-  dout(10) << "share_state peer has fc " << peer_first_committed << " lc " << peer_last_committed << dendl;
+  dout(10) << "share_state peer has fc " << peer_first_committed 
+	   << " lc " << peer_last_committed << dendl;
   version_t v = peer_last_committed + 1;
 
   // start with a stashed full copy?
@@ -169,16 +173,15 @@ void Paxos::share_state(MMonPaxos *m, version_t peer_first_committed, version_t
     bufferlist bl;
     version_t l = get_stashed(bl);
     assert(l <= last_committed);
-    dout(10) << "share_state starting with latest " << l << " (" << bl.length() << " bytes)" << dendl;
+    dout(10) << "share_state starting with latest " << l 
+	     << " (" << bl.length() << " bytes)" << dendl;
     m->latest_value.claim(bl);
     m->latest_version = l;
     v = l;
   }
 
   // include incrementals
-  for ( ;
-       v <= last_committed;
-       v++) {
+  for ( ; v <= last_committed; v++) {
     if (mon->store->exists_bl_sn(machine_name, v)) {
       mon->store->get_bl_sn(m->values[v], machine_name, v);
       dout(10) << " sharing " << v << " (" 
@@ -195,7 +198,8 @@ void Paxos::store_state(MMonPaxos *m)
 
   // stash?
   if (m->latest_version && m->latest_version > last_committed) {
-    dout(10) << "store_state got stash version " << m->latest_version << ", zapping old states" << dendl;
+    dout(10) << "store_state got stash version " 
+	     << m->latest_version << ", zapping old states" << dendl;
 
     assert(start != m->values.end() && start->first == m->latest_version);
 
@@ -221,13 +225,12 @@ void Paxos::store_state(MMonPaxos *m)
     start = m->values.end();
   }
 
-  while (start != m->values.end() &&
-	 start->first <= last_committed)
+  while (start != m->values.end() && start->first <= last_committed) {
     ++start;
+  }
 
   map<version_t,bufferlist>::iterator end = start;
-  while (end != m->values.end() &&
-	 end->first <= m->last_committed) {
+  while (end != m->values.end() && end->first <= m->last_committed) {
     last_committed = end->first;
     if (!first_committed)
       first_committed = last_committed;
@@ -237,7 +240,9 @@ void Paxos::store_state(MMonPaxos *m)
   if (start == end) {
     dout(10) << "store_state nothing to commit" << dendl;
   } else {
-    dout(10) << "store_state [" << start->first << ".." << last_committed << "]" << dendl;
+    dout(10) << "store_state [" << start->first << ".." 
+	     << last_committed << "]" << dendl;
+
     mon->store->put_bl_sn_map(machine_name, start, end);
     mon->store->put_int(last_committed, machine_name, "last_committed");
     mon->store->put_int(first_committed, machine_name, "first_committed");
@@ -304,7 +309,8 @@ void Paxos::handle_last(MMonPaxos *last)
 	if (p->second < last_committed) {
 	  // share committed values
 	  dout(10) << " sending commit to mon." << p->first << dendl;
-	  MMonPaxos *commit = new MMonPaxos(mon->get_epoch(), MMonPaxos::OP_COMMIT, machine_id,
+	  MMonPaxos *commit = new MMonPaxos(mon->get_epoch(),
+					    MMonPaxos::OP_COMMIT, machine_id,
 					    ceph_clock_now(g_ceph_context));
 	  share_state(commit, peer_first_committed[p->first], p->second);
 	  mon->messenger->send_message(commit, mon->monmap->get_inst(p->first));
@@ -381,19 +387,18 @@ void Paxos::begin(bufferlist& v)
     finish_contexts(g_ceph_context, waiting_for_commit);
     finish_contexts(g_ceph_context, waiting_for_readable);
     finish_contexts(g_ceph_context, waiting_for_writeable);
-    update_observers();
     return;
   }
 
-  // ask others to accept it to!
+  // ask others to accept it too!
   for (set<int>::const_iterator p = mon->get_quorum().begin();
        p != mon->get_quorum().end();
        ++p) {
     if (*p == mon->rank) continue;
     
     dout(10) << " sending begin to mon." << *p << dendl;
-    MMonPaxos *begin = new MMonPaxos(mon->get_epoch(), MMonPaxos::OP_BEGIN, machine_id,
-				     ceph_clock_now(g_ceph_context));
+    MMonPaxos *begin = new MMonPaxos(mon->get_epoch(), MMonPaxos::OP_BEGIN,
+				     machine_id, ceph_clock_now(g_ceph_context));
     begin->values[last_committed+1] = new_value;
     begin->last_committed = last_committed;
     begin->pn = accepted_pn;
@@ -430,8 +435,8 @@ void Paxos::handle_begin(MMonPaxos *begin)
   mon->store->put_bl_sn(begin->values[v], machine_name, v);
   
   // reply
-  MMonPaxos *accept = new MMonPaxos(mon->get_epoch(), MMonPaxos::OP_ACCEPT, machine_id,
-				    ceph_clock_now(g_ceph_context));
+  MMonPaxos *accept = new MMonPaxos(mon->get_epoch(), MMonPaxos::OP_ACCEPT,
+				    machine_id, ceph_clock_now(g_ceph_context));
   accept->pn = accepted_pn;
   accept->last_committed = last_committed;
   mon->messenger->send_message(accept, begin->get_source_inst());
@@ -527,8 +532,8 @@ void Paxos::commit()
     if (*p == mon->rank) continue;
 
     dout(10) << " sending commit to mon." << *p << dendl;
-    MMonPaxos *commit = new MMonPaxos(mon->get_epoch(), MMonPaxos::OP_COMMIT, machine_id,
-				      ceph_clock_now(g_ceph_context));
+    MMonPaxos *commit = new MMonPaxos(mon->get_epoch(), MMonPaxos::OP_COMMIT,
+				      machine_id, ceph_clock_now(g_ceph_context));
     commit->values[last_committed] = new_value;
     commit->pn = accepted_pn;
     commit->last_committed = last_committed;
@@ -541,8 +546,6 @@ void Paxos::commit()
 }
 
 
-
-
 void Paxos::handle_commit(MMonPaxos *commit)
 {
   dout(10) << "handle_commit on " << commit->last_committed << dendl;
@@ -571,15 +574,16 @@ void Paxos::extend_lease()
   acked_lease.clear();
   acked_lease.insert(mon->rank);
 
-  dout(7) << "extend_lease now+" << g_conf->mon_lease << " (" << lease_expire << ")" << dendl;
+  dout(7) << "extend_lease now+" << g_conf->mon_lease 
+	  << " (" << lease_expire << ")" << dendl;
 
   // bcast
   for (set<int>::const_iterator p = mon->get_quorum().begin();
-       p != mon->get_quorum().end();
-       ++p) {
+      p != mon->get_quorum().end(); ++p) {
+
     if (*p == mon->rank) continue;
-    MMonPaxos *lease = new MMonPaxos(mon->get_epoch(), MMonPaxos::OP_LEASE, machine_id,
-				     ceph_clock_now(g_ceph_context));
+    MMonPaxos *lease = new MMonPaxos(mon->get_epoch(), MMonPaxos::OP_LEASE,
+				     machine_id, ceph_clock_now(g_ceph_context));
     lease->last_committed = last_committed;
     lease->lease_timestamp = lease_expire;
     lease->first_committed = first_committed;
@@ -590,7 +594,8 @@ void Paxos::extend_lease()
   //  if old timeout is still in place, leave it.
   if (!lease_ack_timeout_event) {
     lease_ack_timeout_event = new C_LeaseAckTimeout(this);
-    mon->timer.add_event_after(g_conf->mon_lease_ack_timeout, lease_ack_timeout_event);
+    mon->timer.add_event_after(g_conf->mon_lease_ack_timeout, 
+			       lease_ack_timeout_event);
   }
 
   // set renew event
@@ -626,7 +631,8 @@ void Paxos::handle_lease(MMonPaxos *lease)
   // sanity
   if (!mon->is_peon() ||
       last_committed != lease->last_committed) {
-    dout(10) << "handle_lease i'm not a peon, or they're not the leader, or the last_committed doesn't match, dropping" << dendl;
+    dout(10) << "handle_lease i'm not a peon, or they're not the leader,"
+	     << " or the last_committed doesn't match, dropping" << dendl;
     lease->put();
     return;
   }
@@ -636,16 +642,23 @@ void Paxos::handle_lease(MMonPaxos *lease)
   // extend lease
   if (lease_expire < lease->lease_timestamp) {
     lease_expire = lease->lease_timestamp;
+
+    // check for extreme clock skew.
+    utime_t now = ceph_clock_now(g_ceph_context);
+    if (lease_expire < now) {
+      utime_t diff = now - lease_expire;
+      derr << "lease_expire from " << lease->get_source_inst() << " is " << diff << " seconds in the past; mons are laggy or clocks are too skewed" << dendl;
+    }
   }
-  
+
   state = STATE_ACTIVE;
   
   dout(10) << "handle_lease on " << lease->last_committed
 	   << " now " << lease_expire << dendl;
 
   // ack
-  MMonPaxos *ack = new MMonPaxos(mon->get_epoch(), MMonPaxos::OP_LEASE_ACK, machine_id,
-				 ceph_clock_now(g_ceph_context));
+  MMonPaxos *ack = new MMonPaxos(mon->get_epoch(), MMonPaxos::OP_LEASE_ACK,
+				 machine_id, ceph_clock_now(g_ceph_context));
   ack->last_committed = last_committed;
   ack->first_committed = first_committed;
   ack->lease_timestamp = ceph_clock_now(g_ceph_context);
@@ -673,7 +686,8 @@ void Paxos::handle_lease_ack(MMonPaxos *ack)
   int from = ack->get_source().num();
 
   if (!lease_ack_timeout_event) {
-    dout(10) << "handle_lease_ack from " << ack->get_source() << " -- stray (probably since revoked)" << dendl;
+    dout(10) << "handle_lease_ack from " << ack->get_source() 
+	     << " -- stray (probably since revoked)" << dendl;
   }
   else if (acked_lease.count(from) == 0) {
     acked_lease.insert(from);
@@ -726,7 +740,6 @@ void Paxos::lease_renew_timeout()
 }
 
 
-
 /*
  * trim old states
  */
@@ -891,72 +904,8 @@ void Paxos::dispatch(PaxosServiceMessage *m)
   default:
     assert(0);
   }
-
-  if (is_readable())
-    update_observers();
 }
 
-void Paxos::register_observer(entity_inst_t inst, version_t v)
-{
-  dout(10) << "register_observer " << inst << " v" << v << dendl;
-  
-  Observer *observer;
-  if (observers.count(inst))
-    observer = observers[inst];
-  else {
-    observers[inst] = observer = new Observer(inst, v);
-  }  
-
-  utime_t timeout = ceph_clock_now(g_ceph_context);
-  timeout += g_conf->paxos_observer_timeout;
-  observer->timeout = timeout;
-
-  if (is_readable())
-    update_observers();
-}
-
-
-void Paxos::update_observers()
-{
-  dout(10) << "update_observers" << dendl;
-
-  bufferlist bl;
-  version_t ver;
-
-  map<entity_inst_t, Observer *>::iterator iter = observers.begin();
-  while (iter != observers.end()) {
-    Observer *observer = iter->second;
-
-    // timed out?
-    if (ceph_clock_now(g_ceph_context) > observer->timeout) {
-      delete observer;
-      observers.erase(iter++);
-      continue;
-    }
-    ++iter;
-    
-    if (observer->last_version == 0 ||
-	observer->last_version < first_committed) {
-      ver = get_stashed(bl);
-      if (ver) {
-	dout(10) << " sending summary state v" << ver << " to " << observer->inst << dendl;
-	mon->messenger->send_message(new MMonObserveNotify(mon->monmap->fsid, machine_id, bl, ver, true),
-				     observer->inst);
-	observer->last_version = ver;
-	continue;
-      }
-    }
-    
-    for (ver = observer->last_version + 1; ver <= last_committed; ver++) {
-      if (read(ver, bl)) {
-	dout(10) << " sending state v" << ver << " to " << observer->inst << dendl;
-	mon->messenger->send_message(new MMonObserveNotify(mon->monmap->fsid, machine_id, bl, ver, false),
-				     observer->inst);
-	observer->last_version = ver;
-      }
-    }
-  }
-}
 
 // -----------------
 // service interface
