@@ -16,6 +16,19 @@
 #ifndef CEPH_FILESTORE_H
 #define CEPH_FILESTORE_H
 
+#include "include/types.h"
+
+#include <map>
+#include <deque>
+#include <boost/scoped_ptr.hpp>
+#include <fstream>
+using namespace std;
+
+#include <ext/hash_map>
+using namespace __gnu_cxx;
+
+#include "include/assert.h"
+
 #include "ObjectStore.h"
 #include "JournalingObjectStore.h"
 
@@ -30,14 +43,6 @@
 
 #include "include/uuid.h"
 
-#include <map>
-#include <deque>
-#include <boost/scoped_ptr.hpp>
-#include <fstream>
-using namespace std;
-
-#include <ext/hash_map>
-using namespace __gnu_cxx;
 
 // from include/linux/falloc.h:
 #ifndef FALLOC_FL_PUNCH_HOLE
@@ -47,7 +52,9 @@ using namespace __gnu_cxx;
 class FileStore : public JournalingObjectStore,
                   public md_config_obs_t
 {
-  static const uint32_t on_disk_version = 2;
+public:
+  static const uint32_t on_disk_version = 3;
+private:
   string internal_name;         ///< internal name, used to name the perfcounter instance
   string basedir, journalpath;
   std::string current_fn;
@@ -83,7 +90,10 @@ class FileStore : public JournalingObjectStore,
   // helper fns
   int get_cdir(coll_t cid, char *s, int len);
   
-  int read_fsid(int fd);
+  /// read a uuid from fd
+  int read_fsid(int fd, uuid_d *uuid);
+
+  /// lock fsid_fd
   int lock_fsid();
 
   // sync thread
@@ -277,10 +287,10 @@ public:
   int lfn_open(coll_t cid, const hobject_t& oid, int flags, mode_t mode);
   int lfn_open(coll_t cid, const hobject_t& oid, int flags);
   int lfn_link(coll_t c, coll_t cid, const hobject_t& o) ;
-  int lfn_unlink(coll_t cid, const hobject_t& o);
+  int lfn_unlink(coll_t cid, const hobject_t& o, const SequencerPosition &spos);
 
  public:
-  FileStore(const std::string &base, const std::string &jdev, const char *internal_name = "filestore");
+  FileStore(const std::string &base, const std::string &jdev, const char *internal_name = "filestore", bool update_to=false);
   ~FileStore();
 
   int _test_fiemap();
@@ -296,7 +306,6 @@ public:
   int mount();
   int umount();
   int get_max_object_name_length();
-  int wipe_subvol(const char *s);
   int mkfs();
   int mkjournal();
 
@@ -305,8 +314,6 @@ public:
   int do_transactions(list<Transaction*> &tls, uint64_t op_seq);
   unsigned apply_transaction(Transaction& t, Context *ondisk=0);
   unsigned apply_transactions(list<Transaction*>& tls, Context *ondisk=0);
-  int _transaction_start(uint64_t bytes, uint64_t ops);
-  void _transaction_finish(int id);
   unsigned _do_transaction(Transaction& t, uint64_t op_seq, int trans_num);
 
   int queue_transaction(Sequencer *osr, Transaction* t);
@@ -324,7 +331,10 @@ public:
    * @param fd open file descriptor for the file/object
    * @param spos sequencer position of the last operation we should not replay
    */
-  void _set_replay_guard(int fd, const SequencerPosition& spos, bool in_progress=false);
+  void _set_replay_guard(int fd,
+			 const SequencerPosition& spos,
+			 const hobject_t *hoid=0,
+			 bool in_progress=false);
 
   /// close a replay guard opened with in_progress=true
   void _close_replay_guard(int fd, const SequencerPosition& spos);
@@ -370,7 +380,7 @@ public:
 		   const SequencerPosition& spos);
   int _do_clone_range(int from, int to, uint64_t srcoff, uint64_t len, uint64_t dstoff);
   int _do_copy_range(int from, int to, uint64_t srcoff, uint64_t len, uint64_t dstoff);
-  int _remove(coll_t cid, const hobject_t& oid);
+  int _remove(coll_t cid, const hobject_t& oid, const SequencerPosition &spos);
 
   void _start_sync();
 
@@ -399,9 +409,12 @@ public:
   int _getattr(const char *fn, const char *name, bufferptr& bp);
   int _getattrs(const char *fn, map<string,bufferptr>& aset, bool user_only = false);
 
-  int _setattrs(coll_t cid, const hobject_t& oid, map<string,bufferptr>& aset);
-  int _rmattr(coll_t cid, const hobject_t& oid, const char *name);
-  int _rmattrs(coll_t cid, const hobject_t& oid);
+  int _setattrs(coll_t cid, const hobject_t& oid, map<string,bufferptr>& aset,
+		const SequencerPosition &spos);
+  int _rmattr(coll_t cid, const hobject_t& oid, const char *name,
+	      const SequencerPosition &spos);
+  int _rmattrs(coll_t cid, const hobject_t& oid,
+	       const SequencerPosition &spos);
 
   int collection_getattr(coll_t c, const char *name, void *value, size_t size);
   int collection_getattr(coll_t c, const char *name, bufferlist& bl);
@@ -410,6 +423,8 @@ public:
   int _collection_setattr(coll_t c, const char *name, const void *value, size_t size);
   int _collection_rmattr(coll_t c, const char *name);
   int _collection_setattrs(coll_t cid, map<string,bufferptr> &aset);
+  int _collection_remove_recursive(const coll_t &cid,
+				   const SequencerPosition &spos);
   int _collection_rename(const coll_t &cid, const coll_t &ncid,
 			 const SequencerPosition& spos);
 
@@ -439,8 +454,6 @@ public:
   int _destroy_collection(coll_t c);
   int _collection_add(coll_t c, coll_t ocid, const hobject_t& o,
 		      const SequencerPosition& spos);
-  int _collection_remove(coll_t c, const hobject_t& o);
-
   void dump_start(const std::string& file);
   void dump_stop();
   void dump_transactions(list<ObjectStore::Transaction*>& ls, uint64_t seq, OpSequencer *osr);
@@ -449,18 +462,21 @@ private:
   void _inject_failure();
 
   // omap
-  int _omap_clear(coll_t cid, const hobject_t &hoid);
+  int _omap_clear(coll_t cid, const hobject_t &hoid,
+		  const SequencerPosition &spos);
   int _omap_setkeys(coll_t cid, const hobject_t &hoid,
-		    const map<string, bufferlist> &aset);
-  int _omap_rmkeys(coll_t cid, const hobject_t &hoid, const set<string> &keys);
-  int _omap_setheader(coll_t cid, const hobject_t &hoid, const bufferlist &bl);
+		    const map<string, bufferlist> &aset,
+		    const SequencerPosition &spos);
+  int _omap_rmkeys(coll_t cid, const hobject_t &hoid, const set<string> &keys,
+		   const SequencerPosition &spos);
+  int _omap_setheader(coll_t cid, const hobject_t &hoid, const bufferlist &bl,
+		      const SequencerPosition &spos);
 
   virtual const char** get_tracked_conf_keys() const;
   virtual void handle_conf_change(const struct md_config_t *conf,
 			  const std::set <std::string> &changed);
   bool m_filestore_btrfs_clone_range;
   bool m_filestore_btrfs_snap;
-  bool m_filestore_btrfs_trans;
   float m_filestore_commit_timeout;
   bool m_filestore_fiemap;
   bool m_filestore_flusher;
@@ -468,13 +484,14 @@ private:
   bool m_filestore_journal_parallel;
   bool m_filestore_journal_trailing;
   bool m_filestore_journal_writeahead;
-  std::string m_filestore_dev;
   int m_filestore_fiemap_threshold;
   bool m_filestore_sync_flush;
   int m_filestore_flusher_max_fds;
+  int m_filestore_flush_min;
   double m_filestore_max_sync_interval;
   double m_filestore_min_sync_interval;
-  bool m_filestore_update_collections;
+  bool m_filestore_fail_eio;
+  int do_update;
   bool m_journal_dio, m_journal_aio;
   std::string m_osd_rollback_to_cluster_snap;
   bool m_osd_use_stale_snap;
