@@ -11,8 +11,6 @@
  * Foundation.  See file COPYING.
  * 
  */
-
-
 #ifndef CEPH_OBJECTSTORE_H
 #define CEPH_OBJECTSTORE_H
 
@@ -20,6 +18,8 @@
 #include "include/buffer.h"
 #include "include/types.h"
 #include "osd/osd_types.h"
+#include "common/TrackedOp.h"
+#include "ObjectMap.h"
 
 #include <errno.h>
 #include <sys/stat.h>
@@ -53,6 +53,8 @@ enum {
   l_os_jq_bytes,
   l_os_j_bytes,
   l_os_j_lat,
+  l_os_j_wr,
+  l_os_j_wr_bytes,
   l_os_oq_max_ops,
   l_os_oq_ops,
   l_os_ops,
@@ -78,35 +80,32 @@ public:
 
   Logger *logger;
 
-  class FragmentationStat {
-  public:
-    int total;
-    int num_extent;
-    int avg_extent;
-    map<int,int> extent_dist;          // powers of two
-    map<int,int> extent_dist_sum;          // powers of two
-
-    float avg_extent_per_object;
-    int avg_extent_jump;  // avg distance bweteen consecutive extents
-
-    int total_free;
-    int num_free_extent;
-    int avg_free_extent;
-    map<int,int> free_extent_dist;     // powers of two
-    map<int,int> free_extent_dist_sum;     // powers of two
-  };
-  
-
+  /**
+   * a sequencer orders transactions
+   *
+   * Any transactions queued under a given sequencer will be applied in
+   * sequence.  Transactions queued under different sequencers may run
+   * in parallel.
+   */
   struct Sequencer_impl {
     virtual void flush() = 0;
     virtual ~Sequencer_impl() {}
   };
   struct Sequencer {
+    string name;
     Sequencer_impl *p;
-    Sequencer() : p(NULL) {}
+
+    Sequencer(string n)
+      : name(n), p(NULL) {}
     ~Sequencer() {
       delete p;
     }
+
+    /// return a unique string identifier for this sequencer
+    const string& get_name() const {
+      return name;
+    }
+    /// wait for any queued transactions on this sequencer to apply
     void flush() {
       if (p)
 	p->flush();
@@ -119,33 +118,41 @@ public:
    */
   class Transaction {
   public:
-    static const int OP_NOP =          0;
-    static const int OP_TOUCH =        9;   // cid, oid
-    static const int OP_WRITE =        10;  // cid, oid, offset, len, bl
-    static const int OP_ZERO =         11;  // cid, oid, offset, len
-    static const int OP_TRUNCATE =     12;  // cid, oid, len
-    static const int OP_REMOVE =       13;  // cid, oid
-    static const int OP_SETATTR =      14;  // cid, oid, attrname, bl
-    static const int OP_SETATTRS =     15;  // cid, oid, attrset
-    static const int OP_RMATTR =       16;  // cid, oid, attrname
-    static const int OP_CLONE =        17;  // cid, oid, newoid
-    static const int OP_CLONERANGE =   18;  // cid, oid, newoid, offset, len
-    static const int OP_CLONERANGE2 =  30;  // cid, oid, newoid, srcoff, len, dstoff
+    enum {
+      OP_NOP =          0,
+      OP_TOUCH =        9,   // cid, oid
+      OP_WRITE =        10,  // cid, oid, offset, len, bl
+      OP_ZERO =         11,  // cid, oid, offset, len
+      OP_TRUNCATE =     12,  // cid, oid, len
+      OP_REMOVE =       13,  // cid, oid
+      OP_SETATTR =      14,  // cid, oid, attrname, bl
+      OP_SETATTRS =     15,  // cid, oid, attrset
+      OP_RMATTR =       16,  // cid, oid, attrname
+      OP_CLONE =        17,  // cid, oid, newoid
+      OP_CLONERANGE =   18,  // cid, oid, newoid, offset, len
+      OP_CLONERANGE2 =  30,  // cid, oid, newoid, srcoff, len, dstoff
 
-    static const int OP_TRIMCACHE =    19;  // cid, oid, offset, len
+      OP_TRIMCACHE =    19,  // cid, oid, offset, len  **DEPRECATED**
 
-    static const int OP_MKCOLL =       20;  // cid
-    static const int OP_RMCOLL =       21;  // cid
-    static const int OP_COLL_ADD =     22;  // cid, oid
-    static const int OP_COLL_REMOVE =  23;  // cid, oid
-    static const int OP_COLL_SETATTR = 24;  // cid, attrname, bl
-    static const int OP_COLL_RMATTR =  25;  // cid, attrname
-    static const int OP_COLL_SETATTRS = 26;  // cid, attrset
+      OP_MKCOLL =       20,  // cid
+      OP_RMCOLL =       21,  // cid
+      OP_COLL_ADD =     22,  // cid, oldcid, oid
+      OP_COLL_REMOVE =  23,  // cid, oid
+      OP_COLL_SETATTR = 24,  // cid, attrname, bl
+      OP_COLL_RMATTR =  25,  // cid, attrname
+      OP_COLL_SETATTRS = 26,  // cid, attrset
+      OP_COLL_MOVE =    8,   // newcid, oldcid, oid
 
-    static const int OP_STARTSYNC =    27;  // start a sync 
+      OP_STARTSYNC =    27,  // start a sync 
 
-    static const int OP_RMATTRS =      28;  // cid, oid
-    static const int OP_COLL_RENAME =       29;  // cid, newcid
+      OP_RMATTRS =      28,  // cid, oid
+      OP_COLL_RENAME =       29,  // cid, newcid
+
+      OP_OMAP_CLEAR = 31,   // cid
+      OP_OMAP_SETKEYS = 32, // cid, attrset
+      OP_OMAP_RMKEYS = 33,  // cid, keyset
+      OP_OMAP_SETHEADER = 34, // cid, header
+    };
 
   private:
     uint64_t ops;
@@ -153,8 +160,13 @@ public:
     uint32_t largest_data_len, largest_data_off, largest_data_off_in_tbl;
     bufferlist tbl;
     bool sobject_encoding;
+    int64_t pool_override;
+    bool use_pool_override;
 
   public:
+    void set_pool_override(int64_t pool) {
+      pool_override = pool;
+    }
 
     void swap(Transaction& other) {
       std::swap(ops, other.ops);
@@ -220,10 +232,14 @@ public:
     class iterator {
       bufferlist::iterator p;
       bool sobject_encoding;
+      int64_t pool_override;
+      bool use_pool_override;
 
       iterator(Transaction *t)
 	: p(t->tbl.begin()),
-	  sobject_encoding(t->sobject_encoding) {}
+	  sobject_encoding(t->sobject_encoding),
+	  pool_override(t->pool_override),
+	  use_pool_override(t->use_pool_override) {}
 
       friend class Transaction;
 
@@ -248,6 +264,10 @@ public:
 	  hoid.oid = soid.oid;
 	} else {
 	  ::decode(hoid, p);
+	  if (use_pool_override && pool_override != -1 &&
+	      hoid.pool == -1) {
+	    hoid.pool = pool_override;
+	  }
 	}
 	return hoid;
       }
@@ -269,12 +289,17 @@ public:
       void get_attrset(map<string,bufferptr>& aset) {
 	::decode(aset, p);
       }
+      void get_attrset(map<string,bufferlist>& aset) {
+	::decode(aset, p);
+      }
+      void get_keyset(set<string> &keys) {
+	::decode(keys, p);
+      }
     };
 
     iterator begin() {
       return iterator(this);
     }
-
     // -----------------------------
 
     void start_sync() {
@@ -319,15 +344,6 @@ public:
       ::encode(len, tbl);
       ops++;
     }
-    void trim_from_cache(coll_t cid, const hobject_t& oid, uint64_t off, uint64_t len) {
-      __u32 op = OP_TRIMCACHE;
-       ::encode(op, tbl);
-      ::encode(cid, tbl);
-      ::encode(oid, tbl);
-      ::encode(off, tbl);
-      ::encode(len, tbl);
-      ops++;
-    }
     void truncate(coll_t cid, const hobject_t& oid, uint64_t off) {
       __u32 op = OP_TRUNCATE;
       ::encode(op, tbl);
@@ -430,6 +446,12 @@ public:
       ::encode(oid, tbl);
       ops++;
     }
+    void collection_move(coll_t cid, coll_t oldcid, const hobject_t& oid) {
+      collection_add(cid, oldcid, oid);
+      collection_remove(oldcid, oid);
+      return;
+    }
+
     void collection_setattr(coll_t cid, const char* name, bufferlist& val) {
       string n(name);
       collection_setattr(cid, n, val);
@@ -469,25 +491,76 @@ public:
       ops++;
     }
 
+    /// Remove omap from hoid
+    void omap_clear(
+      coll_t cid,           ///< [in] Collection containing hoid
+      const hobject_t &hoid ///< [in] Object from which to remove omap
+      ) {
+      __u32 op = OP_OMAP_CLEAR;
+      ::encode(op, tbl);
+      ::encode(cid, tbl);
+      ::encode(hoid, tbl);
+      ops++;
+    }
+    /// Set keys on hoid omap.  Replaces duplicate keys.
+    void omap_setkeys(
+      coll_t cid,                           ///< [in] Collection containing hoid
+      const hobject_t &hoid,                ///< [in] Object to update
+      const map<string, bufferlist> &attrset ///< [in] Replacement keys and values
+      ) {
+      __u32 op = OP_OMAP_SETKEYS;
+      ::encode(op, tbl);
+      ::encode(cid, tbl);
+      ::encode(hoid, tbl);
+      ::encode(attrset, tbl);
+      ops++;
+    }
+    /// Remove keys from hoid omap
+    void omap_rmkeys(
+      coll_t cid,             ///< [in] Collection containing hoid
+      const hobject_t &hoid,  ///< [in] Object from which to remove the omap
+      const set<string> &keys ///< [in] Keys to clear
+      ) {
+      __u32 op = OP_OMAP_RMKEYS;
+      ::encode(op, tbl);
+      ::encode(cid, tbl);
+      ::encode(hoid, tbl);
+      ::encode(keys, tbl);
+      ops++;
+    }
+
+    /// Set omap header
+    void omap_setheader(
+      coll_t cid,             ///< [in] Collection containing hoid
+      const hobject_t &hoid,  ///< [in] Object from which to remove the omap
+      const bufferlist &bl    ///< [in] Header value
+      ) {
+      __u32 op = OP_OMAP_SETHEADER;
+      ::encode(op, tbl);
+      ::encode(cid, tbl);
+      ::encode(hoid, tbl);
+      ::encode(bl, tbl);
+      ops++;
+    }
 
     // etc.
     Transaction() :
       ops(0), pad_unused_bytes(0), largest_data_len(0), largest_data_off(0), largest_data_off_in_tbl(0),
-      sobject_encoding(false) {}
+      sobject_encoding(false), pool_override(-1), use_pool_override(false) {}
     Transaction(bufferlist::iterator &dp) :
       ops(0), pad_unused_bytes(0), largest_data_len(0), largest_data_off(0), largest_data_off_in_tbl(0),
-      sobject_encoding(false) {
+      sobject_encoding(false), pool_override(-1), use_pool_override(false) {
       decode(dp);
     }
     Transaction(bufferlist &nbl) :
       ops(0), pad_unused_bytes(0), largest_data_len(0), largest_data_off(0), largest_data_off_in_tbl(0),
-      sobject_encoding(false) {
+      sobject_encoding(false), pool_override(-1), use_pool_override(false) {
       bufferlist::iterator dp = nbl.begin();
       decode(dp); 
     }
 
     void encode(bufferlist& bl) const {
-      ENCODE_START(5, 5, bl);
+      ENCODE_START(6, 5, bl);
       ::encode(ops, bl);
       ::encode(pad_unused_bytes, bl);
       ::encode(largest_data_len, bl);
@@ -497,7 +570,7 @@ public:
       ENCODE_FINISH(bl);
     }
     void decode(bufferlist::iterator &bl) {
-      DECODE_START_LEGACY_COMPAT_LEN(5, 5, 5, bl);
+      DECODE_START_LEGACY_COMPAT_LEN(6, 5, 5, bl);
       DECODE_OLDEST(2);
       if (struct_v < 4)
 	sobject_encoding = true;
@@ -512,9 +585,11 @@ public:
       }
       ::decode(tbl, bl);
       DECODE_FINISH(bl);
+      if (struct_v < 6) {
+	use_pool_override = true;
+      }
     }
 
-    void dump(ostream& out);
     void dump(ceph::Formatter *f);
     static void generate_test_instances(list<Transaction*>& o);
   };
@@ -526,22 +601,31 @@ public:
       delete t;
     }
   };
-
+  template<class T>
+  struct C_DeleteTransactionHolder : public Context {
+    ObjectStore::Transaction *t;
+    T obj;
+    C_DeleteTransactionHolder(ObjectStore::Transaction *tt, T &obj) :
+      t(tt), obj(obj) {}
+    void finish(int r) {
+      delete t;
+    }
+  };
 
   virtual unsigned apply_transaction(Transaction& t, Context *ondisk=0) = 0;
   virtual unsigned apply_transactions(list<Transaction*>& tls, Context *ondisk=0) = 0;
 
   virtual int queue_transaction(Sequencer *osr, Transaction* t) = 0;
   virtual int queue_transaction(Sequencer *osr, Transaction *t, Context *onreadable, Context *ondisk=0,
-				Context *onreadable_sync=0) {
+				Context *onreadable_sync=0,
+				TrackedOpRef op = TrackedOpRef()) {
     list<Transaction*> tls;
     tls.push_back(t);
-    return queue_transactions(osr, tls, onreadable, ondisk, onreadable_sync);
+    return queue_transactions(osr, tls, onreadable, ondisk, onreadable_sync, op);
   }
   virtual int queue_transactions(Sequencer *osr, list<Transaction*>& tls, Context *onreadable, Context *ondisk=0,
-				 Context *onreadable_sync=0) = 0;
-
-
+				 Context *onreadable_sync=0,
+				 TrackedOpRef op = TrackedOpRef()) = 0;
 
  public:
   ObjectStore() : logger(NULL) {}
@@ -579,23 +663,6 @@ public:
   virtual int read(coll_t cid, const hobject_t& oid, uint64_t offset, size_t len, bufferlist& bl) = 0;
   virtual int fiemap(coll_t cid, const hobject_t& oid, uint64_t offset, size_t len, bufferlist& bl) = 0;
 
-  /*
-  virtual int _remove(coll_t cid, hobject_t oid) = 0;
-  virtual int _truncate(coll_t cid, hobject_t oid, uint64_t size) = 0;
-  virtual int _write(coll_t cid, hobject_t oid, uint64_t offset, size_t len, const bufferlist& bl) = 0;
-  virtual int _zero(coll_t cid, hobject_t oid, uint64_t offset, size_t len) {
-    // write zeros.. yuck!
-    bufferptr bp(len);
-    bufferlist bl;
-    bl.push_back(bp);
-    return _write(cid, oid, offset, len, bl);
-  }
-  */
-
-  virtual void trim_from_cache(coll_t cid, const hobject_t& oid, uint64_t offset, size_t len) = 0; //{ }
-  virtual int is_cached(coll_t cid, const hobject_t& oid, uint64_t offset, size_t len) = 0;  //{ return -1; }
-
-  virtual int getattr(coll_t cid, const hobject_t& oid, const char *name, void *value, size_t size) = 0;
   virtual int getattr(coll_t cid, const hobject_t& oid, const char *name, bufferptr& value) = 0;
   int getattr(coll_t cid, const hobject_t& oid, const char *name, bufferlist& value) {
     bufferptr bp;
@@ -606,18 +673,7 @@ public:
   }
   virtual int getattrs(coll_t cid, const hobject_t& oid, map<string,bufferptr>& aset, bool user_only = false) {return 0;};
 
-  /*
-  virtual int _setattr(coll_t cid, hobject_t oid, const char *name, const void *value, size_t size) = 0;
-  virtual int _setattr(coll_t cid, hobject_t oid, const char *name, const bufferptr &bp) {
-    return _setattr(cid, oid, name, bp.c_str(), bp.length());
-  }
-  virtual int _setattrs(coll_t cid, hobject_t oid, map<string,bufferptr>& aset) = 0;
-  virtual int _rmattr(coll_t cid, hobject_t oid, const char *name) {return 0;}
-  virtual int _clone(coll_t cid, hobject_t oid, hobject_t noid) {
-    return -1;
-  }
-  */
-    
+   
   // collections
   virtual int list_collections(vector<coll_t>& ls) = 0;
   virtual int collection_version_current(coll_t c, uint32_t *version) { 
@@ -632,7 +688,6 @@ public:
   virtual bool collection_empty(coll_t c) = 0;
   virtual int collection_list(coll_t c, vector<hobject_t>& o) = 0;
 
-
   /**
    * list partial contents of collection relative to a hash offset/position
    *
@@ -649,17 +704,58 @@ public:
 				      int min, int max, snapid_t snap, 
 				      vector<hobject_t> *ls, hobject_t *next) = 0;
 
+  /// OMAP
+  /// Get omap contents
+  virtual int omap_get(
+    coll_t c,                ///< [in] Collection containing hoid
+    const hobject_t &hoid,   ///< [in] Object containing omap
+    bufferlist *header,      ///< [out] omap header
+    map<string, bufferlist> *out /// < [out] Key to value map
+    ) = 0;
 
-  /*
-  virtual int _create_collection(coll_t c) = 0;
-  virtual int _destroy_collection(coll_t c) = 0;
-  virtual int _collection_add(coll_t c, coll_t ocid, hobject_t o) = 0;
-  virtual int _collection_remove(coll_t c, hobject_t o) = 0;
-  virtual int _collection_setattr(coll_t cid, const char *name,
-                                 const void *value, size_t size) = 0;
-  virtual int _collection_rmattr(coll_t cid, const char *name) = 0;
-  virtual int _collection_setattrs(coll_t cid, map<string,bufferptr> &aset) = 0;
-  */
+  /// Get omap header
+  virtual int omap_get_header(
+    coll_t c,                ///< [in] Collection containing hoid
+    const hobject_t &hoid,   ///< [in] Object containing omap
+    bufferlist *header       ///< [out] omap header
+    ) = 0;
+
+  /// Get keys defined on hoid
+  virtual int omap_get_keys(
+    coll_t c,              ///< [in] Collection containing hoid
+    const hobject_t &hoid, ///< [in] Object containing omap
+    set<string> *keys      ///< [out] Keys defined on hoid
+    ) = 0;
+
+  /// Get key values
+  virtual int omap_get_values(
+    coll_t c,                    ///< [in] Collection containing hoid
+    const hobject_t &hoid,       ///< [in] Object containing omap
+    const set<string> &keys,     ///< [in] Keys to get
+    map<string, bufferlist> *out ///< [out] Returned keys and values
+    ) = 0;
+
+  /// Filters keys into out which are defined on hoid
+  virtual int omap_check_keys(
+    coll_t c,                ///< [in] Collection containing hoid
+    const hobject_t &hoid,   ///< [in] Object containing omap
+    const set<string> &keys, ///< [in] Keys to check
+    set<string> *out         ///< [out] Subset of keys defined on hoid
+    ) = 0;
+
+  /**
+   * Returns an object map iterator
+   *
+   * Warning!  The returned iterator is an implicit lock on filestore
+   * operations in c.  Do not use filestore methods on c while the returned
+   * iterator is live.  (Filling in a transaction is no problem).
+   *
+   * @return iterator, null on error
+   */
+  virtual ObjectMap::ObjectMapIterator get_omap_iterator(
+    coll_t c,              ///< [in] collection
+    const hobject_t &hoid  ///< [in] object
+    ) = 0;
 
   virtual void sync(Context *onsync) {}
   virtual void sync() {}
@@ -670,13 +766,13 @@ public:
 
   virtual int snapshot(const string& name) { return -EOPNOTSUPP; }
     
-  virtual void _fake_writes(bool b) {};
-  virtual void _get_frag_stat(FragmentationStat& st) {};
-
+  virtual void set_fsid(uuid_d u) = 0;
   virtual uuid_d get_fsid() = 0;
 };
 
 
 WRITE_CLASS_ENCODER(ObjectStore::Transaction)
 
+ostream& operator<<(ostream& out, const ObjectStore::Sequencer& s);
+
 #endif
