@@ -20,6 +20,8 @@
 
 #include "common/code_environment.h"
 
+#define dout_subsys ceph_subsys_osd
+
 // ----------------------------------
 // osd_info_t
 
@@ -84,6 +86,46 @@ ostream& operator<<(ostream& out, const osd_info_t& info)
 // ----------------------------------
 // OSDMap::Incremental
 
+int OSDMap::Incremental::get_net_marked_out(const OSDMap *previous) const
+{
+  int n = 0;
+  for (map<int32_t,uint32_t>::const_iterator p = new_weight.begin();
+       p != new_weight.end();
+       ++p) {
+    if (p->second == CEPH_OSD_OUT && !previous->is_out(p->first))
+      n++;  // marked out
+    if (p->second != CEPH_OSD_OUT && previous->is_out(p->first))
+      n--;  // marked in
+  }
+  return n;
+}
+
+int OSDMap::Incremental::get_net_marked_down(const OSDMap *previous) const
+{
+  int n = 0;
+  for (map<int32_t,uint8_t>::const_iterator p = new_state.begin();
+       p != new_state.end();
+       ++p) {
+    if (p->second & CEPH_OSD_UP) {
+      if (previous->is_up(p->first))
+	n++;  // marked down
+      else
+	n--;  // marked up
+    }
+  }
+  return n;
+}
+
+int OSDMap::Incremental::identify_osd(uuid_d u) const
+{
+  for (map<int32_t,uuid_d>::const_iterator p = new_uuid.begin();
+       p != new_uuid.end();
+       ++p)
+    if (p->second == u)
+      return p->first;
+  return -1;
+}
+
 void OSDMap::Incremental::encode_client_old(bufferlist& bl) const
 {
   __u16 v = 5;
@@ -166,7 +208,7 @@ void OSDMap::Incremental::encode(bufferlist& bl, uint64_t features) const
   ::encode(new_pg_temp, bl);
 
   // extended
-  __u16 ev = CEPH_OSDMAP_INC_VERSION_EXT;
+  __u16 ev = 8;
   ::encode(ev, bl);
   ::encode(new_hb_up, bl);
   ::encode(new_up_thru, bl);
@@ -176,6 +218,7 @@ void OSDMap::Incremental::encode(bufferlist& bl, uint64_t features) const
   ::encode(old_blacklist, bl);
   ::encode(new_up_internal, bl);
   ::encode(cluster_snapshot, bl);
+  ::encode(new_uuid, bl);
 }
 
 void OSDMap::Incremental::decode(bufferlist::iterator &p)
@@ -243,6 +286,10 @@ void OSDMap::Incremental::decode(bufferlist::iterator &p)
     ::decode(new_pg_temp, p);
   }
 
+  // decode short map, too.
+  if (v == 5 && p.end())
+    return;
+
   // extended
   __u16 ev = 0;
   if (v >= 5)
@@ -259,12 +306,174 @@ void OSDMap::Incremental::decode(bufferlist::iterator &p)
     ::decode(new_up_internal, p);
   if (ev >= 7)
     ::decode(cluster_snapshot, p);
+  if (ev >= 8)
+    ::decode(new_uuid, p);
 }
 
+void OSDMap::Incremental::dump(Formatter *f) const
+{
+  f->dump_int("epoch", epoch);
+  f->dump_stream("fsid") << fsid;
+  f->dump_stream("modified") << modified;
+  f->dump_int("new_pool_max", new_pool_max);
+  f->dump_int("new_flags", new_flags);
+
+  if (fullmap.length()) {
+    f->open_object_section("full_map");
+    OSDMap full;
+    bufferlist fbl = fullmap;  // kludge around constness.
+    bufferlist::iterator p = fbl.begin();
+    full.decode(p);
+    full.dump(f);
+    f->close_section();
+  }
+  if (crush.length()) {
+    f->open_object_section("crush");
+    CrushWrapper c;
+    bufferlist tbl = crush;  // kludge around constness.
+    bufferlist::iterator p = tbl.begin();
+    c.decode(p);
+    c.dump(f);
+    f->close_section();
+  }
+
+  f->dump_int("new_max_osd", new_max_osd);
+
+  f->open_array_section("new_pools");
+  for (map<int64_t,pg_pool_t>::const_iterator p = new_pools.begin(); p != new_pools.end(); ++p) {
+    f->open_object_section("pool");
+    f->dump_int("pool", p->first);
+    f->dump_string("name", new_pool_names.find(p->first)->second);
+    p->second.dump(f);
+    f->close_section();
+  }
+  f->close_section();
+  f->open_array_section("old_pools");
+  for (set<int64_t>::const_iterator p = old_pools.begin(); p != old_pools.end(); ++p)
+    f->dump_int("pool", *p);
+  f->close_section();
+
+  f->open_array_section("new_up_osds");
+  for (map<int32_t,entity_addr_t>::const_iterator p = new_up_client.begin(); p != new_up_client.end(); ++p) {
+    f->open_object_section("osd");
+    f->dump_int("osd", p->first);
+    f->dump_stream("public_addr") << p->second;
+    f->dump_stream("cluster_addr") << new_up_internal.find(p->first)->second;
+    f->dump_stream("heartbeat_addr") << new_hb_up.find(p->first)->second;
+    f->close_section();
+  }
+  f->close_section();
+
+  f->open_array_section("new_weight");
+  for (map<int32_t,uint32_t>::const_iterator p = new_weight.begin(); p != new_weight.end(); ++p) {
+    f->open_object_section("osd");
+    f->dump_int("osd", p->first);
+    f->dump_int("weight", p->second);
+    f->close_section();
+  }
+  f->close_section();
+
+  f->open_array_section("osd_state_xor");
+  for (map<int32_t,uint8_t>::const_iterator p = new_state.begin(); p != new_state.end(); ++p) {
+    f->open_object_section("osd");
+    f->dump_int("osd", p->first);
+    set<string> st;
+    calc_state_set(new_state.find(p->first)->second, st);
+    f->open_array_section("state_xor");
+    for (set<string>::iterator p = st.begin(); p != st.end(); ++p)
+      f->dump_string("state", *p);
+    f->close_section();
+  }
+  f->close_section();
+
+  f->open_array_section("new_pg_temp");
+  for (map<pg_t,vector<int> >::const_iterator p = new_pg_temp.begin();
+       p != new_pg_temp.end();
+       p++) {
+    f->open_object_section("pg");
+    f->dump_stream("pgid") << p->first;
+    f->open_array_section("osds");
+    for (vector<int>::const_iterator q = p->second.begin(); q != p->second.end(); ++q)
+      f->dump_int("osd", *q);
+    f->close_section();
+    f->close_section();    
+  }
+  f->close_section();
+
+  f->open_array_section("new_up_thru");
+  for (map<int32_t,uint32_t>::const_iterator p = new_up_thru.begin(); p != new_up_thru.end(); ++p) {
+    f->open_object_section("osd");
+    f->dump_int("osd", p->first);
+    f->dump_int("up_thru", p->second);
+    f->close_section();
+  }
+  f->close_section();
+
+  f->open_array_section("new_lost");
+  for (map<int32_t,uint32_t>::const_iterator p = new_lost.begin(); p != new_lost.end(); ++p) {
+    f->open_object_section("osd");
+    f->dump_int("osd", p->first);
+    f->dump_int("epoch_lost", p->second);
+    f->close_section();
+  }
+  f->close_section();
+
+  f->open_array_section("new_last_clean_interval");
+  for (map<int32_t,pair<epoch_t,epoch_t> >::const_iterator p = new_last_clean_interval.begin();
+       p != new_last_clean_interval.end();
+       ++p) {
+    f->open_object_section("osd");
+    f->dump_int("osd", p->first);
+    f->dump_int("first", p->second.first);
+    f->dump_int("last", p->second.second);
+    f->close_section();
+  }
+  f->close_section();
+
+  f->open_array_section("new_blacklist");
+  for (map<entity_addr_t,utime_t>::const_iterator p = new_blacklist.begin();
+       p != new_blacklist.end();
+       p++) {
+    stringstream ss;
+    ss << p->first;
+    f->dump_stream(ss.str().c_str()) << p->second;
+  }
+  f->close_section();
+  f->open_array_section("old_blacklist");
+  for (vector<entity_addr_t>::const_iterator p = old_blacklist.begin(); p != old_blacklist.end(); ++p)
+    f->dump_stream("addr") << *p;
+  f->close_section();
+
+  if (cluster_snapshot.size())
+    f->dump_string("cluster_snapshot", cluster_snapshot);
+
+  f->open_array_section("new_uuid");
+  for (map<int32_t,uuid_d>::const_iterator p = new_uuid.begin(); p != new_uuid.end(); ++p) {
+    f->open_object_section("osd");
+    f->dump_int("osd", p->first);
+    f->dump_stream("uuid") << p->second;
+    f->close_section();
+  }
+  f->close_section();
+}
+
+void OSDMap::Incremental::generate_test_instances(list<Incremental*>& o)
+{
+  o.push_back(new Incremental);
+}
 
 // ----------------------------------
 // OSDMap
 
+void OSDMap::set_epoch(epoch_t e)
+{
+  epoch = e;
+  for (map<int64_t,pg_pool_t>::iterator p = pools.begin();
+       p != pools.end();
+       p++)
+    p->second.last_change = e;
+}
+
 bool OSDMap::is_blacklisted(const entity_addr_t& a) const
 {
   if (blacklist.empty())
@@ -292,14 +501,14 @@ void OSDMap::set_max_osd(int m)
     osd_weight[o] = CEPH_OSD_OUT;
   }
   osd_info.resize(m);
-  osd_addr.resize(m);
-  osd_cluster_addr.resize(m);
-  osd_hb_addr.resize(m);
+  osd_addrs->client_addr.resize(m);
+  osd_addrs->cluster_addr.resize(m);
+  osd_addrs->hb_addr.resize(m);
+  osd_uuid->resize(m);
 
   calc_num_osds();
 }
 
-
 int OSDMap::calc_num_osds()
 {
   num_osd = 0;
@@ -309,6 +518,42 @@ int OSDMap::calc_num_osds()
   return num_osd;
 }
 
+void OSDMap::get_all_osds(set<int32_t>& ls) const
+{
+  for (int i=0; i<max_osd; i++)
+    if (exists(i))
+      ls.insert(i);
+}
+
+int OSDMap::get_num_up_osds() const
+{
+  int n = 0;
+  for (int i=0; i<max_osd; i++)
+    if (osd_state[i] & CEPH_OSD_EXISTS &&
+	osd_state[i] & CEPH_OSD_UP) n++;
+  return n;
+}
+
+int OSDMap::get_num_in_osds() const
+{
+  int n = 0;
+  for (int i=0; i<max_osd; i++)
+    if (osd_state[i] & CEPH_OSD_EXISTS &&
+	get_weight(i) != CEPH_OSD_OUT) n++;
+  return n;
+}
+
+void OSDMap::calc_state_set(int state, set<string>& st)
+{
+  unsigned t = state;
+  for (unsigned s = 1; t; s <<= 1) {
+    if (t & s) {
+      t &= ~s;
+      st.insert(ceph_osd_state_name(s));
+    }
+  }
+}
+
 void OSDMap::adjust_osd_weights(const map<int,double>& weights, Incremental& inc) const
 {
   float max = 0;
@@ -324,6 +569,82 @@ void OSDMap::adjust_osd_weights(const map<int,double>& weights, Incremental& inc
   }
 }
 
+int OSDMap::identify_osd(const entity_addr_t& addr) const
+{
+  for (int i=0; i<max_osd; i++)
+    if (exists(i) && (get_addr(i) == addr || get_cluster_addr(i) == addr))
+      return i;
+  return -1;
+}
+
+int OSDMap::identify_osd(const uuid_d& u) const
+{
+  for (int i=0; i<max_osd; i++)
+    if (exists(i) && get_uuid(i) == u)
+      return i;
+  return -1;
+}
+
+bool OSDMap::find_osd_on_ip(const entity_addr_t& ip) const
+{
+  for (int i=0; i<max_osd; i++)
+    if (exists(i) && (get_addr(i).is_same_host(ip) || get_cluster_addr(i).is_same_host(ip)))
+      return i;
+  return -1;
+}
+
+void OSDMap::dedup(const OSDMap *o, OSDMap *n)
+{
+  if (o->epoch == n->epoch)
+    return;
+
+  int diff = 0;
+
+  // do addrs match?
+  if (o->max_osd != n->max_osd)
+    diff++;
+  for (int i = 0; i < o->max_osd && i < n->max_osd; i++) {
+    if ( n->osd_addrs->client_addr[i] &&  o->osd_addrs->client_addr[i] &&
+	*n->osd_addrs->client_addr[i] == *o->osd_addrs->client_addr[i])
+      n->osd_addrs->client_addr[i] = o->osd_addrs->client_addr[i];
+    else
+      diff++;
+    if ( n->osd_addrs->cluster_addr[i] &&  o->osd_addrs->cluster_addr[i] &&
+	*n->osd_addrs->cluster_addr[i] == *o->osd_addrs->cluster_addr[i])
+      n->osd_addrs->cluster_addr[i] = o->osd_addrs->cluster_addr[i];
+    else
+      diff++;
+    if ( n->osd_addrs->hb_addr[i] &&  o->osd_addrs->hb_addr[i] &&
+	*n->osd_addrs->hb_addr[i] == *o->osd_addrs->hb_addr[i])
+      n->osd_addrs->hb_addr[i] = o->osd_addrs->hb_addr[i];
+    else
+      diff++;
+  }
+  if (diff == 0) {
+    // zoinks, no differences at all!
+    n->osd_addrs = o->osd_addrs;
+  }
+
+  // does crush match?
+  bufferlist oc, nc;
+  ::encode(*o->crush, oc);
+  ::encode(*n->crush, nc);
+  if (oc.contents_equal(nc)) {
+    n->crush = o->crush;
+  }
+
+  // does pg_temp match?
+  if (o->pg_temp->size() == n->pg_temp->size()) {
+    if (*o->pg_temp == *n->pg_temp)
+      n->pg_temp = o->pg_temp;
+  }
+
+  // do uuids match?
+  if (o->osd_uuid->size() == n->osd_uuid->size() &&
+      *o->osd_uuid == *n->osd_uuid)
+    n->osd_uuid = o->osd_uuid;
+}
+
 int OSDMap::apply_incremental(Incremental &inc)
 {
   if (inc.epoch == 1)
@@ -367,6 +688,8 @@ int OSDMap::apply_incremental(Incremental &inc)
   for (map<int64_t,string>::iterator p = inc.new_pool_names.begin();
        p != inc.new_pool_names.end();
        p++) {
+    if (pool_name.count(p->first))
+      name_pool.erase(pool_name[p->first]);
     pool_name[p->first] = p->second;
     name_pool[p->second] = p->first;
   }
@@ -390,23 +713,27 @@ int OSDMap::apply_incremental(Incremental &inc)
 	(s & CEPH_OSD_UP)) {
       osd_info[i->first].down_at = epoch;
     }
+    if ((osd_state[i->first] & CEPH_OSD_EXISTS) &&
+	(s & CEPH_OSD_EXISTS))
+      (*osd_uuid)[i->first] = uuid_d();
     osd_state[i->first] ^= s;
   }
   for (map<int32_t,entity_addr_t>::iterator i = inc.new_up_client.begin();
        i != inc.new_up_client.end();
        i++) {
     osd_state[i->first] |= CEPH_OSD_EXISTS | CEPH_OSD_UP;
-    osd_addr[i->first] = i->second;
+    osd_addrs->client_addr[i->first].reset(new entity_addr_t(i->second));
     if (inc.new_hb_up.empty())
-      osd_hb_addr[i->first] = i->second;	//this is a backward-compatibility hack
+      osd_addrs->hb_addr[i->first].reset(new entity_addr_t(i->second)); //this is a backward-compatibility hack
     else
-      osd_hb_addr[i->first] = inc.new_hb_up[i->first];
+      osd_addrs->hb_addr[i->first].reset(new entity_addr_t(inc.new_hb_up[i->first]));
     osd_info[i->first].up_from = epoch;
   }
   for (map<int32_t,entity_addr_t>::iterator i = inc.new_up_internal.begin();
        i != inc.new_up_internal.end();
        i++)
-    osd_cluster_addr[i->first] = i->second;
+    osd_addrs->cluster_addr[i->first].reset(new entity_addr_t(i->second));
+
   // info
   for (map<int32_t,epoch_t>::iterator i = inc.new_up_thru.begin();
        i != inc.new_up_thru.end();
@@ -421,12 +748,16 @@ int OSDMap::apply_incremental(Incremental &inc)
   for (map<int32_t,epoch_t>::iterator p = inc.new_lost.begin(); p != inc.new_lost.end(); p++)
     osd_info[p->first].lost_at = p->second;
 
+  // uuid
+  for (map<int32_t,uuid_d>::iterator p = inc.new_uuid.begin(); p != inc.new_uuid.end(); ++p) 
+    (*osd_uuid)[p->first] = p->second;
+
   // pg rebuild
   for (map<pg_t, vector<int> >::iterator p = inc.new_pg_temp.begin(); p != inc.new_pg_temp.end(); p++) {
     if (p->second.empty())
-      pg_temp.erase(p->first);
+      pg_temp->erase(p->first);
     else
-      pg_temp[p->first] = p->second;
+      (*pg_temp)[p->first] = p->second;
   }
 
   // blacklist
@@ -451,13 +782,161 @@ int OSDMap::apply_incremental(Incremental &inc)
   // do new crush map last (after up/down stuff)
   if (inc.crush.length()) {
     bufferlist::iterator blp = inc.crush.begin();
-    crush.decode(blp);
+    crush.reset(new CrushWrapper);
+    crush->decode(blp);
   }
 
   calc_num_osds();
   return 0;
 }
 
+
+// mapping
+int OSDMap::object_locator_to_pg(const object_t& oid, const object_locator_t& loc, pg_t &pg) const
+{
+  // calculate ps (placement seed)
+  const pg_pool_t *pool = get_pg_pool(loc.get_pool());
+  if (!pool)
+    return -ENOENT;
+  ps_t ps;
+  if (loc.key.length())
+    ps = ceph_str_hash(pool->object_hash, loc.key.c_str(), loc.key.length());
+  else
+    ps = ceph_str_hash(pool->object_hash, oid.name.c_str(), oid.name.length());
+  pg = pg_t(ps, loc.get_pool(), -1);
+  return 0;
+}
+
+ceph_object_layout OSDMap::make_object_layout(object_t oid, int pg_pool) const
+{
+  object_locator_t loc(pg_pool);
+
+  ceph_object_layout ol;
+  pg_t pgid = object_locator_to_pg(oid, loc);
+  ol.ol_pgid = pgid.get_old_pg().v;
+  ol.ol_stripe_unit = 0;
+  return ol;
+}
+
+void OSDMap::_remove_nonexistent_osds(vector<int>& osds) const
+{
+  unsigned removed = 0;
+  for (unsigned i = 0; i < osds.size(); i++) {
+    if (!exists(osds[i])) {
+      removed++;
+      continue;
+    }
+    if (removed) {
+      osds[i - removed] = osds[i];
+    }
+  }
+  if (removed)
+    osds.resize(osds.size() - removed);
+}
+
+int OSDMap::_pg_to_osds(const pg_pool_t& pool, pg_t pg, vector<int>& osds) const
+{
+  // map to osds[]
+  ps_t pps = pool.raw_pg_to_pps(pg);  // placement ps
+  unsigned size = pool.get_size();
+
+  // what crush rule?
+  int ruleno = crush->find_rule(pool.get_crush_ruleset(), pool.get_type(), size);
+  if (ruleno >= 0)
+    crush->do_rule(ruleno, pps, osds, size, osd_weight);
+
+  _remove_nonexistent_osds(osds);
+
+  return osds.size();
+}
+
+// pg -> (up osd list)
+void OSDMap::_raw_to_up_osds(pg_t pg, vector<int>& raw, vector<int>& up) const
+{
+  up.clear();
+  for (unsigned i=0; i<raw.size(); i++) {
+    if (!exists(raw[i]) || is_down(raw[i])) 
+      continue;
+    up.push_back(raw[i]);
+  }
+}
+  
+bool OSDMap::_raw_to_temp_osds(const pg_pool_t& pool, pg_t pg, vector<int>& raw, vector<int>& temp) const
+{
+  pg = pool.raw_pg_to_pg(pg);
+  map<pg_t,vector<int> >::const_iterator p = pg_temp->find(pg);
+  if (p != pg_temp->end()) {
+    temp.clear();
+    for (unsigned i=0; i<p->second.size(); i++) {
+      if (!exists(p->second[i]) || is_down(p->second[i]))
+	continue;
+      temp.push_back(p->second[i]);
+    }
+    return true;
+  }
+  return false;
+}
+
+int OSDMap::pg_to_osds(pg_t pg, vector<int>& raw) const
+{
+  const pg_pool_t *pool = get_pg_pool(pg.pool());
+  if (!pool)
+    return 0;
+  return _pg_to_osds(*pool, pg, raw);
+}
+
+int OSDMap::pg_to_acting_osds(pg_t pg, vector<int>& acting) const
+{
+  const pg_pool_t *pool = get_pg_pool(pg.pool());
+  if (!pool)
+    return 0;
+  vector<int> raw;
+  _pg_to_osds(*pool, pg, raw);
+  if (!_raw_to_temp_osds(*pool, pg, raw, acting))
+    _raw_to_up_osds(pg, raw, acting);
+  return acting.size();
+}
+
+void OSDMap::pg_to_raw_up(pg_t pg, vector<int>& up) const
+{
+  const pg_pool_t *pool = get_pg_pool(pg.pool());
+  if (!pool)
+    return;
+  vector<int> raw;
+  _pg_to_osds(*pool, pg, raw);
+  _raw_to_up_osds(pg, raw, up);
+}
+  
+void OSDMap::pg_to_up_acting_osds(pg_t pg, vector<int>& up, vector<int>& acting) const
+{
+  const pg_pool_t *pool = get_pg_pool(pg.pool());
+  if (!pool)
+    return;
+  vector<int> raw;
+  _pg_to_osds(*pool, pg, raw);
+  _raw_to_up_osds(pg, raw, up);
+  if (!_raw_to_temp_osds(*pool, pg, raw, acting))
+    acting = up;
+}
+
+int OSDMap::calc_pg_rank(int osd, vector<int>& acting, int nrep)
+{
+  if (!nrep)
+    nrep = acting.size();
+  for (int i=0; i<nrep; i++) 
+    if (acting[i] == osd)
+      return i;
+  return -1;
+}
+
+int OSDMap::calc_pg_role(int osd, vector<int>& acting, int nrep)
+{
+  if (!nrep)
+    nrep = acting.size();
+  return calc_pg_rank(osd, acting, nrep);
+}
+
+
 // serialize, unserialize
 void OSDMap::encode_client_old(bufferlist& bl) const
 {
@@ -499,13 +978,13 @@ void OSDMap::encode_client_old(bufferlist& bl) const
   ::encode(max_osd, bl);
   ::encode(osd_state, bl);
   ::encode(osd_weight, bl);
-  ::encode(osd_addr, bl);
+  ::encode(osd_addrs->client_addr, bl);
 
   // for ::encode(pg_temp, bl);
-  n = pg_temp.size();
+  n = pg_temp->size();
   ::encode(n, bl);
-  for (map<pg_t,vector<int32_t> >::const_iterator p = pg_temp.begin();
-       p != pg_temp.end();
+  for (map<pg_t,vector<int32_t> >::const_iterator p = pg_temp->begin();
+       p != pg_temp->end();
        ++p) {
     old_pg_t opg = p->first.get_old_pg();
     ::encode(opg, bl);
@@ -514,7 +993,7 @@ void OSDMap::encode_client_old(bufferlist& bl) const
 
   // crush
   bufferlist cbl;
-  crush.encode(cbl);
+  crush->encode(cbl);
   ::encode(cbl, bl);
 }
 
@@ -543,24 +1022,25 @@ void OSDMap::encode(bufferlist& bl, uint64_t features) const
   ::encode(max_osd, bl);
   ::encode(osd_state, bl);
   ::encode(osd_weight, bl);
-  ::encode(osd_addr, bl);
+  ::encode(osd_addrs->client_addr, bl);
 
-  ::encode(pg_temp, bl);
+  ::encode(*pg_temp, bl);
 
   // crush
   bufferlist cbl;
-  crush.encode(cbl);
+  crush->encode(cbl);
   ::encode(cbl, bl);
 
   // extended
-  __u16 ev = CEPH_OSDMAP_VERSION_EXT;
+  __u16 ev = 8;
   ::encode(ev, bl);
-  ::encode(osd_hb_addr, bl);
+  ::encode(osd_addrs->hb_addr, bl);
   ::encode(osd_info, bl);
   ::encode(blacklist, bl);
-  ::encode(osd_cluster_addr, bl);
+  ::encode(osd_addrs->cluster_addr, bl);
   ::encode(cluster_snapshot_epoch, bl);
   ::encode(cluster_snapshot, bl);
+  ::encode(*osd_uuid, bl);
 }
 
 void OSDMap::decode(bufferlist& bl)
@@ -581,37 +1061,39 @@ void OSDMap::decode(bufferlist::iterator& p)
   ::decode(created, p);
   ::decode(modified, p);
 
-  int32_t max_pools = 0;
-  if (v < 4) {
-    ::decode(max_pools, p);
-  }
   if (v < 6) {
+    if (v < 4) {
+      int32_t max_pools = 0;
+      ::decode(max_pools, p);
+      pool_max = max_pools;
+    }
     pools.clear();
     ::decode(n, p);
     while (n--) {
       ::decode(t, p);
       ::decode(pools[t], p);
     }
+    if (v == 4) {
+      ::decode(n, p);
+      pool_max = n;
+    } else if (v == 5) {
+      pool_name.clear();
+      ::decode(n, p);
+      while (n--) {
+	::decode(t, p);
+	::decode(pool_name[t], p);
+      }
+      ::decode(n, p);
+      pool_max = n;
+    }
   } else {
     ::decode(pools, p);
-  }
-  if (v == 5) {
-    pool_name.clear();
-    ::decode(n, p);
-    while (n--) {
-      ::decode(t, p);
-      ::decode(pool_name[t], p);
-    }
-  } else if (v >= 6) {
     ::decode(pool_name, p);
-  }
-  if (v == 4 || v == 5) {
-    ::decode(n, p);
-    pool_max = n;
-  } else if (v >= 6) {
     ::decode(pool_max, p);
-  } else {
-    pool_max = max_pools;
+  }
+  // kludge around some old bug that zeroed out pool_max (#2307)
+  if (pools.size() && pool_max < pools.rbegin()->first) {
+    pool_max = pools.rbegin()->first;
   }
 
   ::decode(flags, p);
@@ -619,45 +1101,51 @@ void OSDMap::decode(bufferlist::iterator& p)
   ::decode(max_osd, p);
   ::decode(osd_state, p);
   ::decode(osd_weight, p);
-  ::decode(osd_addr, p);
+  ::decode(osd_addrs->client_addr, p);
   if (v <= 5) {
-    pg_temp.clear();
+    pg_temp->clear();
     ::decode(n, p);
     while (n--) {
       old_pg_t opg;
       ::decode_raw(opg, p);
-      ::decode(pg_temp[pg_t(opg)], p);
+      ::decode((*pg_temp)[pg_t(opg)], p);
     }
   } else {
-    ::decode(pg_temp, p);
+    ::decode(*pg_temp, p);
   }
 
   // crush
   bufferlist cbl;
   ::decode(cbl, p);
   bufferlist::iterator cblp = cbl.begin();
-  crush.decode(cblp);
+  crush->decode(cblp);
 
   // extended
   __u16 ev = 0;
   if (v >= 5)
     ::decode(ev, p);
-  ::decode(osd_hb_addr, p);
+  ::decode(osd_addrs->hb_addr, p);
   ::decode(osd_info, p);
   if (v < 5)
     ::decode(pool_name, p);
 
   ::decode(blacklist, p);
   if (ev >= 6)
-    ::decode(osd_cluster_addr, p);
+    ::decode(osd_addrs->cluster_addr, p);
   else
-    osd_cluster_addr.resize(osd_addr.size());
+    osd_addrs->cluster_addr.resize(osd_addrs->client_addr.size());
 
   if (ev >= 7) {
     ::decode(cluster_snapshot_epoch, p);
     ::decode(cluster_snapshot, p);
   }
 
+  if (ev >= 8) {
+    ::decode(*osd_uuid, p);
+  } else {
+    osd_uuid->resize(max_osd);
+  }
+
   // index pool names
   name_pool.clear();
   for (map<int64_t,string>::iterator i = pool_name.begin(); i != pool_name.end(); i++)
@@ -685,6 +1173,7 @@ void OSDMap::dump(Formatter *f) const
   f->dump_stream("modified") << get_modified();
   f->dump_string("flags", get_flag_string());
   f->dump_string("cluster_snapshot", get_cluster_snapshot());
+  f->dump_int("pool_max", get_pool_max());
   f->dump_int("max_osd", get_max_osd());
 
   f->open_array_section("pools");
@@ -701,6 +1190,7 @@ void OSDMap::dump(Formatter *f) const
     if (exists(i)) {
       f->open_object_section("osd_info");
       f->dump_int("osd", i);
+      f->dump_stream("uuid") << get_uuid(i);
       f->dump_int("up", is_up(i));
       f->dump_int("in", is_in(i));
       get_info(i).dump(f);
@@ -720,8 +1210,8 @@ void OSDMap::dump(Formatter *f) const
   f->close_section();
 
   f->open_array_section("pg_temp");
-  for (map<pg_t,vector<int> >::const_iterator p = pg_temp.begin();
-       p != pg_temp.end();
+  for (map<pg_t,vector<int> >::const_iterator p = pg_temp->begin();
+       p != pg_temp->end();
        p++) {
     f->open_array_section("osds");
     for (vector<int>::const_iterator q = p->second.begin(); q != p->second.end(); ++q)
@@ -748,27 +1238,40 @@ void OSDMap::generate_test_instances(list<OSDMap*>& o)
   CephContext *cct = new CephContext(CODE_ENVIRONMENT_UTILITY);
   o.push_back(new OSDMap);
   uuid_d fsid;
-  o.back()->build_simple(cct, 1, fsid, 16, 7, 8, 9);
+  o.back()->build_simple(cct, 1, fsid, 16, 7, 8);
   o.back()->created = o.back()->modified = utime_t(1, 2);  // fix timestamp
-  delete cct;
+  cct->put();
+}
+
+string OSDMap::get_flag_string(unsigned f)
+{
+  string s;
+  if ( f& CEPH_OSDMAP_NEARFULL)
+    s += ",nearfull";
+  if (f & CEPH_OSDMAP_FULL)
+    s += ",full";
+  if (f & CEPH_OSDMAP_PAUSERD)
+    s += ",pauserd";
+  if (f & CEPH_OSDMAP_PAUSEWR)
+    s += ",pausewr";
+  if (f & CEPH_OSDMAP_PAUSEREC)
+    s += ",pauserec";
+  if (f & CEPH_OSDMAP_NOUP)
+    s += ",no-up";
+  if (f & CEPH_OSDMAP_NODOWN)
+    s += ",no-down";
+  if (f & CEPH_OSDMAP_NOOUT)
+    s += ",no-out";
+  if (f & CEPH_OSDMAP_NOIN)
+    s += ",no-in";
+  if (s.length())
+    s = s.erase(0, 1);
+  return s;
 }
 
 string OSDMap::get_flag_string() const
 {
-  string s;
-  if (test_flag(CEPH_OSDMAP_NEARFULL))
-    s += ",nearfull";
-  if (test_flag(CEPH_OSDMAP_FULL))
-    s += ",full";
-  if (test_flag(CEPH_OSDMAP_PAUSERD))
-    s += ",pauserd";
-  if (test_flag(CEPH_OSDMAP_PAUSEWR))
-    s += ",pausewr";
-  if (test_flag(CEPH_OSDMAP_PAUSEREC))
-    s += ",pauserec";
-  if (s.length())
-    s = s.erase(0, 1);
-  return s;
+  return get_flag_string(flags);
 }
 
 struct qi {
@@ -821,13 +1324,15 @@ void OSDMap::print(ostream& out) const
       set<string> st;
       get_state(i, st);
       out << " " << st;
+      if (!get_uuid(i).is_zero())
+	out << " " << get_uuid(i);
       out << "\n";
     }
   }
   out << std::endl;
 
-  for (map<pg_t,vector<int> >::const_iterator p = pg_temp.begin();
-       p != pg_temp.end();
+  for (map<pg_t,vector<int> >::const_iterator p = pg_temp->begin();
+       p != pg_temp->end();
        p++)
     out << "pg_temp " << p->first << " " << p->second << "\n";
 
@@ -858,10 +1363,10 @@ void OSDMap::print_tree(ostream& out) const
   out << "# id\tweight\ttype name\tup/down\treweight\n";
   set<int> touched;
   set<int> roots;
-  crush.find_roots(roots);
+  crush->find_roots(roots);
   for (set<int>::iterator p = roots.begin(); p != roots.end(); p++) {
     list<qi> q;
-    q.push_back(qi(*p, 0, crush.get_bucket_weight(*p) / (float)0x10000));
+    q.push_back(qi(*p, 0, crush->get_bucket_weight(*p) / (float)0x10000));
     while (!q.empty()) {
       int cur = q.front().item;
       int depth = q.front().depth;
@@ -879,14 +1384,14 @@ void OSDMap::print_tree(ostream& out) const
 	continue;
       }
 
-      int type = crush.get_bucket_type(cur);
-      out << crush.get_type_name(type) << " " << crush.get_item_name(cur) << "\n";
+      int type = crush->get_bucket_type(cur);
+      out << crush->get_type_name(type) << " " << crush->get_item_name(cur) << "\n";
 
       // queue bucket contents...
-      int s = crush.get_bucket_size(cur);
+      int s = crush->get_bucket_size(cur);
       for (int k=s-1; k>=0; k--)
-	q.push_front(qi(crush.get_bucket_item(cur, k), depth+1,
-			(float)crush.get_bucket_item_weight(cur, k) / (float)0x10000));
+	q.push_front(qi(crush->get_bucket_item(cur, k), depth+1,
+			(float)crush->get_bucket_item_weight(cur, k) / (float)0x10000));
     }
   }
 
@@ -919,11 +1424,11 @@ void OSDMap::print_summary(ostream& out) const
 }
 
 void OSDMap::build_simple(CephContext *cct, epoch_t e, uuid_d &fsid,
-			  int nosd, int pg_bits, int pgp_bits, int lpg_bits)
+			  int nosd, int pg_bits, int pgp_bits)
 {
   ldout(cct, 10) << "build_simple on " << num_osd
 		 << " osds with " << pg_bits << " pg bits per osd, "
-		 << lpg_bits << " lpg bits" << dendl;
+		 << dendl;
   epoch = e;
   set_fsid(fsid);
   created = modified = ceph_clock_now(cct);
@@ -948,17 +1453,16 @@ void OSDMap::build_simple(CephContext *cct, epoch_t e, uuid_d &fsid,
     pools[pool].size = cct->_conf->osd_pool_default_size;
     pools[pool].crush_ruleset = p->first;
     pools[pool].object_hash = CEPH_STR_HASH_RJENKINS;
-    pools[pool].pg_num = poolbase << pg_bits;
-    pools[pool].pgp_num = poolbase << pgp_bits;
-    pools[pool].lpg_num = lpg_bits ? (1 << (lpg_bits-1)) : 0;
-    pools[pool].lpgp_num = lpg_bits ? (1 << (lpg_bits-1)) : 0;
+    pools[pool].set_pg_num(poolbase << pg_bits);
+    pools[pool].set_pgp_num(poolbase << pgp_bits);
     pools[pool].last_change = epoch;
     if (p->first == CEPH_DATA_RULE)
       pools[pool].crash_replay_interval = cct->_conf->osd_default_data_pool_replay_window;
     pool_name[pool] = p->second;
+    name_pool[p->second] = pool;
   }
 
-  build_simple_crush_map(cct, crush, rulesets, nosd);
+  build_simple_crush_map(cct, *crush, rulesets, nosd);
 
   for (int i=0; i<nosd; i++) {
     set_state(i, 0);
@@ -977,10 +1481,13 @@ void OSDMap::build_simple_crush_map(CephContext *cct, CrushWrapper& crush,
   crush.set_type_name(0, "osd");
   crush.set_type_name(1, "host");
   crush.set_type_name(2, "rack");
-  crush.set_type_name(3, "pool");
+  crush.set_type_name(3, "row");
+  crush.set_type_name(4, "room");
+  crush.set_type_name(5, "datacenter");
+  crush.set_type_name(6, "pool");
 
   // root
-  int rootid = crush.add_bucket(0, CRUSH_BUCKET_STRAW, CRUSH_HASH_DEFAULT, 3 /* pool */, 0, NULL, NULL);
+  int rootid = crush.add_bucket(0, CRUSH_BUCKET_STRAW, CRUSH_HASH_DEFAULT, 6 /* pool */, 0, NULL, NULL);
   crush.set_item_name(rootid, "default");
 
   for (int o=0; o<nosd; o++) {
@@ -1001,6 +1508,7 @@ void OSDMap::build_simple_crush_map(CephContext *cct, CrushWrapper& crush,
   for (map<int,const char*>::iterator p = rulesets.begin(); p != rulesets.end(); p++) {
     int ruleset = p->first;
     crush_rule *rule = crush_make_rule(3, ruleset, pg_pool_t::TYPE_REP, minrep, maxrep);
+    assert(rule);
     crush_rule_set_step(rule, 0, CRUSH_RULE_TAKE, rootid, 0);
     // just spread across osds
     crush_rule_set_step(rule, 1, CRUSH_RULE_CHOOSE_FIRSTN, CRUSH_CHOOSE_N, 0);
@@ -1012,12 +1520,12 @@ void OSDMap::build_simple_crush_map(CephContext *cct, CrushWrapper& crush,
   crush.finalize();
 }
 
-void OSDMap::build_simple_from_conf(CephContext *cct, epoch_t e, uuid_d &fsid,
-				    int pg_bits, int pgp_bits, int lpg_bits)
+int OSDMap::build_simple_from_conf(CephContext *cct, epoch_t e, uuid_d &fsid,
+				   int pg_bits, int pgp_bits)
 {
   ldout(cct, 10) << "build_simple_from_conf with "
 		 << pg_bits << " pg bits per osd, "
-		 << lpg_bits << " lpg bits" << dendl;
+		 << dendl;
   epoch = e;
   set_fsid(fsid);
   created = modified = ceph_clock_now(cct);
@@ -1025,7 +1533,7 @@ void OSDMap::build_simple_from_conf(CephContext *cct, epoch_t e, uuid_d &fsid,
   const md_config_t *conf = cct->_conf;
 
   // count osds
-  int maxosd = 0;
+  int maxosd = 0, numosd = 0;
 
   vector<string> sections;
   conf->get_all_sections(sections);
@@ -1039,6 +1547,11 @@ void OSDMap::build_simple_from_conf(CephContext *cct, epoch_t e, uuid_d &fsid,
     if (*end != '\0')
       continue;
 
+    if (o > cct->_conf->mon_max_osd) {
+      lderr(cct) << "[osd." << o << "] in config has id > mon_max_osd " << cct->_conf->mon_max_osd << dendl;
+      return -ERANGE;
+    }
+    numosd++;
     if (o > maxosd)
       maxosd = o;
   }
@@ -1061,22 +1574,23 @@ void OSDMap::build_simple_from_conf(CephContext *cct, epoch_t e, uuid_d &fsid,
     pools[pool].size = cct->_conf->osd_pool_default_size;
     pools[pool].crush_ruleset = p->first;
     pools[pool].object_hash = CEPH_STR_HASH_RJENKINS;
-    pools[pool].pg_num = (maxosd + 1) << pg_bits;
-    pools[pool].pgp_num = (maxosd + 1) << pgp_bits;
-    pools[pool].lpg_num = lpg_bits ? (1 << (lpg_bits-1)) : 0;
-    pools[pool].lpgp_num = lpg_bits ? (1 << (lpg_bits-1)) : 0;
+    pools[pool].set_pg_num((numosd + 1) << pg_bits);
+    pools[pool].set_pgp_num((numosd + 1) << pgp_bits);
     pools[pool].last_change = epoch;
     if (p->first == CEPH_DATA_RULE)
       pools[pool].crash_replay_interval = cct->_conf->osd_default_data_pool_replay_window;
     pool_name[pool] = p->second;
+    name_pool[p->second] = pool;
   }
 
-  build_simple_crush_map_from_conf(cct, crush, rulesets);
+  build_simple_crush_map_from_conf(cct, *crush, rulesets);
 
   for (int i=0; i<=maxosd; i++) {
     set_state(i, 0);
     set_weight(i, CEPH_OSD_OUT);
   }
+
+  return 0;
 }
 
 void OSDMap::build_simple_crush_map_from_conf(CephContext *cct, CrushWrapper& crush,
@@ -1089,12 +1603,15 @@ void OSDMap::build_simple_crush_map_from_conf(CephContext *cct, CrushWrapper& cr
   crush.set_type_name(0, "osd");
   crush.set_type_name(1, "host");
   crush.set_type_name(2, "rack");
-  crush.set_type_name(3, "pool");
+  crush.set_type_name(3, "row");
+  crush.set_type_name(4, "room");
+  crush.set_type_name(5, "datacenter");
+  crush.set_type_name(6, "pool");
 
   set<string> hosts, racks;
 
   // root
-  int rootid = crush.add_bucket(0, CRUSH_BUCKET_STRAW, CRUSH_HASH_DEFAULT, 3 /* pool */, 0, NULL, NULL);
+  int rootid = crush.add_bucket(0, CRUSH_BUCKET_STRAW, CRUSH_HASH_DEFAULT, 6 /* pool */, 0, NULL, NULL);
   crush.set_item_name(rootid, "default");
 
   // add osds
@@ -1110,13 +1627,16 @@ void OSDMap::build_simple_crush_map_from_conf(CephContext *cct, CrushWrapper& cr
     if (*end != '\0')
       continue;
 
-    string host;
-    string rack;
+    string host, rack, row, room, dc, pool;
     vector<string> sections;
     sections.push_back("osd");
     sections.push_back(*i);
     conf->get_val_from_conf_file(sections, "host", host, false);
     conf->get_val_from_conf_file(sections, "rack", rack, false);
+    conf->get_val_from_conf_file(sections, "row", row, false);
+    conf->get_val_from_conf_file(sections, "room", room, false);
+    conf->get_val_from_conf_file(sections, "datacenter", dc, false);
+    conf->get_val_from_conf_file(sections, "pool", pool, false);
 
     if (host.length() == 0)
       host = "unknownhost";
@@ -1129,9 +1649,15 @@ void OSDMap::build_simple_crush_map_from_conf(CephContext *cct, CrushWrapper& cr
     map<string,string> loc;
     loc["host"] = host;
     loc["rack"] = rack;
+    if (row.size())
+      loc["row"] = row;
+    if (room.size())
+      loc["room"] = room;
+    if (dc.size())
+      loc["datacenter"] = dc;
     loc["pool"] = "default";
 
-    ldout(cct, 0) << " adding osd." << o << " at " << loc << dendl;
+    ldout(cct, 5) << " adding osd." << o << " at " << loc << dendl;
     crush.insert_item(cct, o, 1.0, *i, loc);
   }
 
@@ -1141,6 +1667,7 @@ void OSDMap::build_simple_crush_map_from_conf(CephContext *cct, CrushWrapper& cr
   for (map<int,const char*>::iterator p = rulesets.begin(); p != rulesets.end(); p++) {
     int ruleset = p->first;
     crush_rule *rule = crush_make_rule(3, ruleset, pg_pool_t::TYPE_REP, minrep, maxrep);
+    assert(rule);
     crush_rule_set_step(rule, 0, CRUSH_RULE_TAKE, rootid, 0);
 
     if (racks.size() > 3) {
