@@ -11,6 +11,7 @@ using namespace std;
 #include "common/Formatter.h"
 #include "global/global_init.h"
 #include "common/errno.h"
+#include "include/utime.h"
 
 #include "common/armor.h"
 #include "rgw_user.h"
@@ -46,6 +47,8 @@ void _usage()
   cerr << "  bucket unlink              unlink bucket from specified user\n";
   cerr << "  bucket stats               returns bucket statistics\n";
   cerr << "  bucket info                show bucket information\n";
+  cerr << "  bucket rm                  remove bucket\n";
+  cerr << "  object rm                  remove object\n";
   cerr << "  pool add                   add an existing pool for data placement\n";
   cerr << "  pool rm                    remove an existing pool from data placement set\n";
   cerr << "  pools list                 list placement active set\n";
@@ -54,6 +57,8 @@ void _usage()
   cerr << "  log show                   dump a log from specific object or (bucket + date\n";
   cerr << "                             + bucket-id)\n";
   cerr << "  log rm                     remove log object\n";
+  cerr << "  usage show                 show usage (by user, date range)\n";
+  cerr << "  usage trim                 trim usage (by user, date range)\n";
   cerr << "  temp remove                remove temporary objects that were created up to\n";
   cerr << "                             specified date (and optional time)\n";
   cerr << "options:\n";
@@ -72,8 +77,9 @@ void _usage()
   cerr << "   --bucket=<bucket>\n";
   cerr << "   --pool=<pool>\n";
   cerr << "   --object=<object>\n";
-  cerr << "   --date=<yyyy-mm-dd>\n";
-  cerr << "   --time=<HH:MM:SS>\n";
+  cerr << "   --date=<date>\n";
+  cerr << "   --start-date=<date>\n";
+  cerr << "   --end-date=<date>\n";
   cerr << "   --bucket-id=<bucket-id>\n";
   cerr << "   --format=<format>         specify output format for certain operations: xml,\n";
   cerr << "                             json\n";
@@ -81,10 +87,18 @@ void _usage()
   cerr << "                             user data\n";
   cerr << "   --purge-keys              when specified, subuser removal will also purge all the\n";
   cerr << "                             subuser keys\n";
+  cerr << "   --purge-objects           remove a bucket's objects before deleting it\n";
+  cerr << "                             (NOTE: required to delete a non-empty bucket)\n";
+  cerr << "   --lazy-remove             defer the removal of the tail of an object until the intent\n";
+  cerr << "                             log is processed.\n";
   cerr << "   --show-log-entries=<flag> enable/disable dump of log entries on log show\n";
   cerr << "   --show-log-sum=<flag>     enable/disable dump of log summation on log show\n";
   cerr << "   --skip-zero-entries       log show only dumps entries that don't have zero value\n";
   cerr << "                             in one of the numeric field\n";
+  cerr << "   --yes-i-really-mean-it    required for certain operations\n";
+  cerr << "\n";
+  cerr << "<date> := \"YYYY-MM-DD[ hh:mm:ss]\"\n";
+  cerr << "\n";
   generic_client_usage();
 }
 
@@ -117,6 +131,7 @@ enum {
   OPT_BUCKET_LINK,
   OPT_BUCKET_UNLINK,
   OPT_BUCKET_STATS,
+  OPT_BUCKET_RM,
   OPT_POLICY,
   OPT_POOL_ADD,
   OPT_POOL_RM,
@@ -124,7 +139,10 @@ enum {
   OPT_LOG_LIST,
   OPT_LOG_SHOW,
   OPT_LOG_RM,
+  OPT_USAGE_SHOW,
+  OPT_USAGE_TRIM,
   OPT_TEMP_REMOVE,
+  OPT_OBJECT_RM,
 };
 
 static uint32_t str_to_perm(const char *str)
@@ -195,6 +213,8 @@ static int get_cmd(const char *cmd, const char *prev_cmd, bool *need_more)
       strcmp(cmd, "pool") == 0 ||
       strcmp(cmd, "pools") == 0 ||
       strcmp(cmd, "log") == 0 ||
+      strcmp(cmd, "usage") == 0 ||
+      strcmp(cmd, "object") == 0 ||
       strcmp(cmd, "temp") == 0) {
     *need_more = true;
     return 0;
@@ -243,6 +263,8 @@ static int get_cmd(const char *cmd, const char *prev_cmd, bool *need_more)
       return OPT_BUCKET_UNLINK;
     if (strcmp(cmd, "stats") == 0)
       return OPT_BUCKET_STATS;
+    if (strcmp(cmd, "rm") == 0)
+      return OPT_BUCKET_RM;
   } else if (strcmp(prev_cmd, "log") == 0) {
     if (strcmp(cmd, "list") == 0)
       return OPT_LOG_LIST;
@@ -250,6 +272,11 @@ static int get_cmd(const char *cmd, const char *prev_cmd, bool *need_more)
       return OPT_LOG_SHOW;
     if (strcmp(cmd, "rm") == 0)
       return OPT_LOG_RM;
+  } else if (strcmp(prev_cmd, "usage") == 0) {
+    if (strcmp(cmd, "show") == 0)
+      return OPT_USAGE_SHOW;
+    if (strcmp(cmd, "trim") == 0)
+      return OPT_USAGE_TRIM;
   } else if (strcmp(prev_cmd, "temp") == 0) {
     if (strcmp(cmd, "remove") == 0)
       return OPT_TEMP_REMOVE;
@@ -261,6 +288,9 @@ static int get_cmd(const char *cmd, const char *prev_cmd, bool *need_more)
   } else if (strcmp(prev_cmd, "pools") == 0) {
     if (strcmp(cmd, "list") == 0)
       return OPT_POOLS_LIST;
+  } else if (strcmp(prev_cmd, "object") == 0) {
+    if (strcmp(cmd, "rm") == 0)
+      return OPT_OBJECT_RM;
   }
 
   return -EINVAL;
@@ -301,6 +331,7 @@ static void show_user_info(RGWUserInfo& info, Formatter *formatter)
   formatter->dump_string("display_name", info.display_name);
   formatter->dump_string("email", info.user_email);
   formatter->dump_int("suspended", (int)info.suspended);
+  formatter->dump_int("max_buckets", (int)info.max_buckets);
 
   // subusers
   formatter->open_array_section("subusers");
@@ -500,6 +531,133 @@ enum ObjectKeyType {
   KEY_TYPE_S3,
 };
 
+static void parse_date(string& date, uint64_t *epoch, string *out_date = NULL, string *out_time = NULL)
+{
+  struct tm tm;
+
+  memset(&tm, 0, sizeof(tm));
+
+  const char *p = strptime(date.c_str(), "%Y-%m-%d", &tm);
+  if (p) {
+    if (*p == ' ') {
+      p++;
+      strptime(p, " %H:%M:%S", &tm);
+    }
+  } else {
+    return;
+  }
+  time_t t = timegm(&tm);
+  if (epoch)
+    *epoch = (uint64_t)t;
+
+  if (out_date) {
+    char buf[32];
+    strftime(buf, sizeof(buf), "%F", &tm);
+    *out_date = buf;
+  }
+  if (out_time) {
+    char buf[32];
+    strftime(buf, sizeof(buf), "%T", &tm);
+    *out_time = buf;
+  }
+}
+
+static int remove_shadow_file_now(void *user_ctx, rgw_obj& obj, RGWIntentEvent intent)
+{
+  int r = rgwstore->delete_obj(NULL,obj);
+  return r;
+}
+
+static int remove_shadow_file_eventually(void *user_ctx, rgw_obj& obj, RGWIntentEvent intent)
+{
+  int r = rgw_log_intent(obj, DEL_OBJ, ceph_clock_now(g_ceph_context),
+                         g_conf->rgw_intent_log_object_name_utc);
+
+  return r;
+}
+
+static int remove_object(rgw_bucket& bucket, std::string& object, bool delete_object_tail_later)
+{
+  int ret = -EINVAL;
+  RGWRadosCtx *rctx = new RGWRadosCtx();
+  rgw_obj obj(bucket,object);
+
+  if (delete_object_tail_later)
+    rgwstore->set_intent_cb(rctx, remove_shadow_file_eventually);
+  else
+    rgwstore->set_intent_cb(rctx, remove_shadow_file_now);
+
+  ret = rgwstore->delete_obj(rctx, obj);
+
+  return ret;
+}
+
+static int remove_bucket(rgw_bucket& bucket, bool delete_children, bool delete_object_tail_later)
+{
+  int ret;
+  map<RGWObjCategory, RGWBucketStats> stats;
+  std::vector<RGWObjEnt> objs;
+  std::string prefix, delim, marker, ns;
+  map<string, bool> common_prefixes;
+  rgw_obj obj;
+  RGWBucketInfo info;
+  bufferlist bl;
+
+  static rgw_bucket pi_buckets_rados = RGW_ROOT_BUCKET;
+  ret = rgwstore->get_bucket_stats(bucket, stats);
+  if (ret < 0)
+    return ret;
+
+  obj.bucket = bucket;
+  int max = 1000;
+
+  ret = rgw_get_obj(NULL, pi_buckets_rados, bucket.name, bl, NULL);
+
+  bufferlist::iterator iter = bl.begin();
+  try {
+    ::decode(info, iter);
+  } catch (buffer::error& err) {
+    cerr << "ERROR: could not decode buffer info, caught buffer::error" << std::endl;
+    return -EIO;
+  }
+
+  if (delete_children) {
+    ret = rgwstore->list_objects(bucket, max, prefix, delim, marker, objs, common_prefixes,
+                                 false, ns, (bool *)false, NULL);
+    if (ret < 0)
+      return ret;
+
+    while (objs.size() > 0) {
+      std::vector<RGWObjEnt>::iterator it = objs.begin();
+      for (it = objs.begin(); it != objs.end(); it++) {
+        ret = remove_object(bucket, (*it).name, delete_object_tail_later);
+        if (ret < 0)
+          return ret;
+      }
+      objs.clear();
+
+      ret = rgwstore->list_objects(bucket, max, prefix, delim, marker, objs, common_prefixes,
+                                   false, ns, (bool *)false, NULL);
+      if (ret < 0)
+        return ret;
+    }
+  }
+
+  ret = rgwstore->delete_bucket(bucket);
+  if (ret < 0) {
+    cerr << "ERROR: could not remove bucket " << bucket.name << std::endl;
+
+    return ret;
+  }
+
+  ret = rgw_remove_user_bucket_info(info.owner, bucket);
+  if (ret < 0) {
+    cerr << "ERROR: unable to remove user bucket information" << std::endl;
+  }
+
+  return ret;
+}
+
 int main(int argc, char **argv) 
 {
   vector<const char*> args;
@@ -511,11 +669,13 @@ int main(int argc, char **argv)
 
   std::string user_id, access_key, secret_key, user_email, display_name;
   std::string bucket_name, pool_name, object;
-  std::string date, time, subuser, access, format;
+  std::string date, subuser, access, format;
+  std::string start_date, end_date;
   std::string key_type_str;
   ObjectKeyType key_type = KEY_TYPE_S3;
   rgw_bucket bucket;
   uint32_t perm_mask = 0;
+  bool specified_perm_mask = false;
   uint64_t auid = -1;
   RGWUserInfo info;
   RGWRados *store;
@@ -537,6 +697,10 @@ int main(int argc, char **argv)
   int show_log_sum = true;
   int skip_zero_entries = false;  // log show
   int purge_keys = false;
+  int yes_i_really_mean_it = false;
+  int delete_child_objects = false;
+  int delete_object_tail_later = false;
+  int max_buckets = -1;
 
   std::string val;
   std::ostringstream errs;
@@ -591,13 +755,20 @@ int main(int argc, char **argv)
 	exit(EXIT_FAILURE);
       }
       auid = tmp;
-    } else if (ceph_argparse_witharg(args, i, &val, "--date", (char*)NULL)) {
+    } else if (ceph_argparse_witharg(args, i, &val, "--max-buckets", (char*)NULL)) {
+      max_buckets = atoi(val.c_str());
+    } else if (ceph_argparse_witharg(args, i, &val, "--date", "--time", (char*)NULL)) {
       date = val;
-    } else if (ceph_argparse_witharg(args, i, &val, "--time", (char*)NULL)) {
-      time = val;
+      if (end_date.empty())
+        end_date = date;
+    } else if (ceph_argparse_witharg(args, i, &val, "--start-date", "--start-time", (char*)NULL)) {
+      start_date = val;
+    } else if (ceph_argparse_witharg(args, i, &val, "--end-date", "--end-time", (char*)NULL)) {
+      end_date = val;
     } else if (ceph_argparse_witharg(args, i, &val, "--access", (char*)NULL)) {
       access = val;
       perm_mask = str_to_perm(access.c_str());
+      specified_perm_mask = true;
     } else if (ceph_argparse_witharg(args, i, &val, "--bucket-id", (char*)NULL)) {
       bucket_id = val;
       if (bucket_id.empty()) {
@@ -606,12 +777,19 @@ int main(int argc, char **argv)
       }
     } else if (ceph_argparse_witharg(args, i, &val, "--format", (char*)NULL)) {
       format = val;
+    } else if (ceph_argparse_binary_flag(args, i, &delete_child_objects, NULL, "--purge-objects", (char*)NULL)) {
+      delete_child_objects = true;
+    } else if (ceph_argparse_binary_flag(args, i, &delete_object_tail_later, NULL, "--lazy-remove", (char*)NULL)) {
+      delete_object_tail_later = true;
     } else if (ceph_argparse_binary_flag(args, i, &pretty_format, NULL, "--pretty-format", (char*)NULL)) {
       // do nothing
     } else if (ceph_argparse_binary_flag(args, i, &purge_data, NULL, "--purge-data", (char*)NULL)) {
-      // do nothing
+      delete_child_objects = true;
+      purge_data = true;
     } else if (ceph_argparse_binary_flag(args, i, &purge_keys, NULL, "--purge-keys", (char*)NULL)) {
       // do nothing
+    } else if (ceph_argparse_binary_flag(args, i, &yes_i_really_mean_it, NULL, "--yes-i-really-mean-it", (char*)NULL)) {
+      // do nothing
     } else {
       ++i;
     }
@@ -761,6 +939,10 @@ int main(int argc, char **argv)
         cerr << "error: subuser already exists" << std::endl;
         return 1;
       }
+      if (!key_type_str.empty() && key_type == KEY_TYPE_S3) {
+        cerr << "error: subusers may not be created with an S3 key, aborting" << std::endl;
+        return 1;
+      }
     } else if (!subuser_found) {
       cerr << "error: subuser doesn't exist" << std::endl;
       return 1;
@@ -805,30 +987,15 @@ int main(int argc, char **argv)
   map<string, RGWSubUser>::iterator uiter;
   RGWUserInfo old_info = info;
 
-  if ((!bucket_name.empty()) || !bucket_id.empty()) {
-    if (!bucket_id.empty()) {
-      int ret = rgwstore->get_bucket_info(NULL, bucket_id, bucket_info);
-
-      if (ret < 0) {
-        cerr << "could not retrieve bucket info for bucket_id=" << bucket_id << std::endl;
-        return ret;
-      }
-      bucket = bucket_info.bucket;
-      if ((!bucket_name.empty()) && bucket.name.compare(bucket_name.c_str()) != 0) {
-        cerr << "bucket name does not match bucket id (expected bucket name: " << bucket.name << ")" << std::endl;
-        return -EINVAL;
-      }
-    } else {
-      string bucket_name_str = bucket_name;
-      RGWBucketInfo bucket_info;
-      int r = rgwstore->get_bucket_info(NULL, bucket_name_str, bucket_info);
-      if (r < 0) {
-        cerr << "could not get bucket info for bucket=" << bucket_name_str << std::endl;
-        return r;
-      }
-      bucket = bucket_info.bucket;
-      bucket_id = bucket.bucket_id;
+  if (!bucket_name.empty()) {
+    string bucket_name_str = bucket_name;
+    RGWBucketInfo bucket_info;
+    int r = rgwstore->get_bucket_info(NULL, bucket_name_str, bucket_info);
+    if (r < 0) {
+      cerr << "could not get bucket info for bucket=" << bucket_name_str << std::endl;
+      return r;
     }
+    bucket = bucket_info.bucket;
   }
 
   int err;
@@ -840,6 +1007,8 @@ int main(int argc, char **argv)
   case OPT_KEY_CREATE:
     if (!user_id.empty())
       info.user_id = user_id;
+    if (max_buckets >= 0)
+      info.max_buckets = max_buckets;
     if (key_type == KEY_TYPE_SWIFT) {
       access_key = info.user_id;
       access_key.append(":");
@@ -873,9 +1042,10 @@ int main(int argc, char **argv)
     if (auid != (uint64_t)-1)
       info.auid = auid;
     if (!subuser.empty()) {
-      RGWSubUser u;
+      RGWSubUser u = info.subusers[subuser];
       u.name = subuser;
-      u.perm_mask = perm_mask;
+      if (specified_perm_mask)
+        u.perm_mask = perm_mask;
 
       info.subusers[subuser] = u;
     }
@@ -1051,7 +1221,9 @@ int main(int argc, char **argv)
       cerr << "date wasn't specified" << std::endl;
       return usage();
     }
-    int r = store->remove_temp_objects(date, time);
+    string parsed_date, parsed_time;
+    parse_date(date, NULL, &parsed_date, &parsed_time);
+    int r = store->remove_temp_objects(parsed_date, parsed_time);
     if (r < 0) {
       cerr << "failure removing temp objects: " << cpp_strerror(r) << std::endl;
       return 1;
@@ -1210,10 +1382,22 @@ next:
   
   if (opt_cmd == OPT_USER_RM) {
     RGWUserBuckets buckets;
+    int ret;
+
     if (rgw_read_user_buckets(user_id, buckets, false) >= 0) {
       map<string, RGWBucketEnt>& m = buckets.get_buckets();
-      if (m.size() > 0) {
-        cerr << "user bucket list not empty, can't remove user" << std::endl;
+
+      if (m.size() > 0 && purge_data) {
+        for (std::map<string, RGWBucketEnt>::iterator it = m.begin(); it != m.end(); it++) {
+          ret = remove_bucket(((*it).second).bucket, true, delete_object_tail_later);
+
+          if (ret < 0)
+            return ret;
+        }
+      }
+
+      if (m.size() > 0 && !purge_data) {
+        cerr << "ERROR: specify --purge-data to remove a user with a non-empty bucket list" << std::endl;
         return 1;
       }
     }
@@ -1263,8 +1447,8 @@ next:
   }
 
   if (opt_cmd == OPT_BUCKET_STATS) {
-    if (bucket_name.empty() && bucket_id.empty() && user_id.empty()) {
-      cerr << "either bucket or bucket-id or uid needs to be specified" << std::endl;
+    if (bucket_name.empty() && user_id.empty()) {
+      cerr << "either bucket or uid needs to be specified" << std::endl;
       return usage();
     }
     formatter->reset();
@@ -1328,5 +1512,141 @@ next:
     }
   } 
 
+  if (opt_cmd == OPT_USAGE_SHOW) {
+    uint64_t start_epoch = 0;
+    uint64_t end_epoch = (uint64_t)-1;
+
+    parse_date(start_date, &start_epoch);
+    parse_date(end_date, &end_epoch);
+
+    uint32_t max_entries = 1000;
+
+    bool is_truncated = true;
+
+    RGWUsageIter usage_iter;
+
+    map<rgw_user_bucket, rgw_usage_log_entry> usage;
+
+    formatter->open_object_section("usage");
+    if (show_log_entries) {
+      formatter->open_array_section("entries");
+    }
+    string last_owner;
+    bool user_section_open = false;
+    map<string, rgw_usage_log_entry> summary_map;
+    while (is_truncated) {
+      int ret = rgwstore->read_usage(user_id, start_epoch, end_epoch, max_entries,
+                                     &is_truncated, usage_iter, usage);
+
+      if (ret == -ENOENT) {
+        ret = 0;
+        is_truncated = false;
+      }
+
+      if (ret < 0) {
+        cerr << "ERROR: read_usage() returned ret=" << ret << std::endl;
+        break;
+      }
+
+      map<rgw_user_bucket, rgw_usage_log_entry>::iterator iter;
+      for (iter = usage.begin(); iter != usage.end(); ++iter) {
+        const rgw_user_bucket& ub = iter->first;
+        const rgw_usage_log_entry& entry = iter->second;
+
+        if (show_log_entries) {
+          if (ub.user.compare(last_owner) != 0) {
+            if (user_section_open) {
+              formatter->close_section();
+              formatter->close_section();
+            }
+            formatter->open_object_section("user");
+            formatter->dump_string("owner", ub.user);
+            formatter->open_array_section("buckets");
+            user_section_open = true;
+            last_owner = ub.user;
+          }
+          formatter->open_object_section("bucket");
+          formatter->dump_string("bucket", ub.bucket);
+          utime_t ut(entry.epoch, 0);
+          ut.gmtime(formatter->dump_stream("time"));
+          formatter->dump_int("epoch", entry.epoch);
+          formatter->dump_int("bytes_sent", entry.bytes_sent);
+          formatter->dump_int("bytes_received", entry.bytes_received);
+          formatter->dump_int("ops", entry.ops);
+          formatter->dump_int("successful_ops", entry.successful_ops);
+          formatter->close_section(); // bucket
+          formatter->flush(cout);
+        }
+
+        summary_map[ub.user].aggregate(entry);
+      }
+    }
+    if (show_log_entries) {
+      if (user_section_open) {
+        formatter->close_section(); // buckets
+        formatter->close_section(); //user
+      }
+      formatter->close_section(); // entries
+    }
+
+    if (show_log_sum) {
+      formatter->open_array_section("summary");
+      map<string, rgw_usage_log_entry>::iterator siter;
+      for (siter = summary_map.begin(); siter != summary_map.end(); ++siter) {
+        const rgw_usage_log_entry& entry = siter->second;
+        formatter->open_object_section("user");
+        formatter->dump_string("user", siter->first);
+        formatter->dump_int("bytes_sent", entry.bytes_sent);
+        formatter->dump_int("bytes_received", entry.bytes_received);
+        formatter->dump_int("ops", entry.ops);
+        formatter->dump_int("successful_ops", entry.successful_ops);
+        formatter->close_section();
+        formatter->flush(cout);
+      }
+
+      formatter->close_section(); // summary
+    }
+
+    formatter->close_section(); // usage
+    formatter->flush(cout);
+  }
+
+  if (opt_cmd == OPT_USAGE_TRIM) {
+    if (user_id.empty() && !yes_i_really_mean_it) {
+      cerr << "usage trim without user specified will remove *all* users data" << std::endl;
+      cerr << "do you really mean it? (requires --yes-i-really-mean-it)" << std::endl;
+      return 1;
+    }
+    uint64_t start_epoch = 0;
+    uint64_t end_epoch = (uint64_t)-1;
+
+    parse_date(start_date, &start_epoch);
+    parse_date(end_date, &end_epoch);
+
+    int ret = rgwstore->trim_usage(user_id, start_epoch, end_epoch);
+    if (ret < 0) {
+      cerr << "ERROR: read_usage() returned ret=" << ret << std::endl;
+      return 1;
+    }   
+  }
+
+  if (opt_cmd == OPT_OBJECT_RM) {
+    int ret = remove_object(bucket, object, delete_object_tail_later);
+
+    if (ret < 0) {
+      cerr << "ERROR: object remove returned: " << cpp_strerror(-ret) << std::endl;
+      return 1;
+    }
+  }
+
+  if (opt_cmd == OPT_BUCKET_RM) {
+    int ret = remove_bucket(bucket, delete_child_objects, delete_object_tail_later);
+
+    if (ret < 0) {
+      cerr << "ERROR: bucket remove returned: " << cpp_strerror(-ret) << std::endl;
+      return 1;
+    }
+  }
+
   return 0;
 }
