@@ -86,7 +86,8 @@ is_err() const
 
 req_state::req_state(CephContext *_cct, struct RGWEnv *e) : cct(_cct), os_auth_token(NULL), os_user(NULL), os_groups(NULL), env(e)
 {
-  should_log = env->conf->should_log;
+  enable_ops_log = env->conf->enable_ops_log;
+  enable_usage_log = env->conf->enable_usage_log;
   content_started = false;
   format = 0;
   bucket_acl = NULL;
@@ -135,10 +136,36 @@ static bool check_str_end(const char *s)
   return true;
 }
 
+static bool check_gmt_end(const char *s)
+{
+  if (!s || !*s)
+    return false;
+
+  while (isspace(*s)) {
+    ++s;
+  }
+
+  /* check for correct timezone */
+  if ((strncmp(s, "GMT", 3) != 0) &&
+      (strncmp(s, "UTC", 3) != 0)) {
+    return false;
+  }
+
+  /* trailing space */
+  s += 3;
+  while (isspace(*s)) {
+    ++s;
+  }
+  if (*s)
+    return false;
+
+  return true;
+}
+
 static bool parse_rfc850(const char *s, struct tm *t)
 {
   memset(t, 0, sizeof(*t));
-  return check_str_end(strptime(s, "%A, %d-%b-%y %H:%M:%S GMT", t));
+  return check_gmt_end(strptime(s, "%A, %d-%b-%y %H:%M:%S ", t));
 }
 
 static bool parse_asctime(const char *s, struct tm *t)
@@ -150,7 +177,7 @@ static bool parse_asctime(const char *s, struct tm *t)
 static bool parse_rfc1123(const char *s, struct tm *t)
 {
   memset(t, 0, sizeof(*t));
-  return check_str_end(strptime(s, "%a, %d %b %Y %H:%M:%S GMT", t));
+  return check_gmt_end(strptime(s, "%a, %d %b %Y %H:%M:%S ", t));
 }
 
 static bool parse_rfc1123_alt(const char *s, struct tm *t)
@@ -293,16 +320,28 @@ int XMLArgs::parse()
     NameVal nv(nameval);
     int ret = nv.parse();
     if (ret >= 0) {
-      val_map[nv.get_name()] = nv.get_val();
+      string& name = nv.get_name();
+      string& val = nv.get_val();
+      val_map[name] = val;
 
-      if ((nv.get_name().compare("acl") == 0) ||
-          (nv.get_name().compare("location") == 0) ||
-          (nv.get_name().compare("uploads") == 0) ||
-          (nv.get_name().compare("partNumber") == 0) ||
-          (nv.get_name().compare("uploadId") == 0) ||
-          (nv.get_name().compare("versionid") == 0) ||
-          (nv.get_name().compare("torrent") == 0)) {
-        sub_resources[nv.get_name()] = nv.get_val();
+      if ((name.compare("acl") == 0) ||
+          (name.compare("location") == 0) ||
+          (name.compare("uploads") == 0) ||
+          (name.compare("partNumber") == 0) ||
+          (name.compare("uploadId") == 0) ||
+          (name.compare("versionId") == 0) ||
+          (name.compare("torrent") == 0)) {
+        sub_resources[name] = val;
+      } else if (name[0] == 'r') { // root of all evil
+        if ((name.compare("response-content-type") == 0) ||
+           (name.compare("response-content-language") == 0) ||
+           (name.compare("response-expires") == 0) ||
+           (name.compare("response-cache-control") == 0) ||
+           (name.compare("response-content-disposition") == 0) ||
+           (name.compare("response-content-encoding") == 0)) {
+	  sub_resources[name] = val;
+	  has_resp_modifier = true;
+	}
       }
     }
 
@@ -312,19 +351,22 @@ int XMLArgs::parse()
   return 0;
 }
 
-string& XMLArgs::get(string& name)
+string& XMLArgs::get(string& name, bool *exists)
 {
   map<string, string>::iterator iter;
   iter = val_map.find(name);
-  if (iter == val_map.end())
-    return empty_str;
-  return iter->second;
+  bool e = (iter != val_map.end());
+  if (exists)
+    *exists = e;
+  if (e)
+    return iter->second;
+  return empty_str;
 }
 
-string& XMLArgs::get(const char *name)
+string& XMLArgs::get(const char *name, bool *exists)
 {
   string s(name);
-  return get(s);
+  return get(s, exists);
 }
 
 bool verify_bucket_permission(struct req_state *s, int perm)
@@ -396,7 +438,7 @@ static char hex_to_num(char c)
 bool url_decode(string& src_str, string& dest_str)
 {
   const char *src = src_str.c_str();
-  char dest[src_str.size()];
+  char dest[src_str.size() + 1];
   int pos = 0;
   char c;
 
