@@ -26,7 +26,7 @@ using namespace librados;
 
 #include "rgw_log.h"
 
-#define DOUT_SUBSYS rgw
+#define dout_subsys ceph_subsys_rgw
 
 using namespace std;
 
@@ -38,7 +38,8 @@ static RGWRados rados_provider;
 RGWRados* RGWRados::store;
 
 
-static string notify_oid = "notify";
+static string notify_oid_prefix = "notify";
+static string *notify_oids = NULL;
 static string shadow_ns = "shadow";
 static string dir_oid_prefix = ".dir.";
 static string default_storage_pool = ".rgw.buckets";
@@ -50,280 +51,10 @@ static rgw_bucket pi_buckets_rados = RGW_ROOT_BUCKET;
 static RGWObjCategory shadow_category = RGW_OBJ_CATEGORY_SHADOW;
 static RGWObjCategory main_category = RGW_OBJ_CATEGORY_MAIN;
 
-void rgw_bucket_pending_info::generate_test_instances(list<rgw_bucket_pending_info*>& o)
-{
-  rgw_bucket_pending_info *i = new rgw_bucket_pending_info;
-  i->state = CLS_RGW_STATE_COMPLETE;
-  i->op = CLS_RGW_OP_DEL;
-  o.push_back(i);
-  o.push_back(new rgw_bucket_pending_info);
-}
+#define RGW_USAGE_OBJ_PREFIX "usage."
 
-void rgw_bucket_pending_info::dump(Formatter *f) const
-{
-  f->dump_int("state", (int)state);
-  f->dump_stream("timestamp") << timestamp;
-  f->dump_int("op", (int)op);
-}
-
-void rgw_bucket_dir_entry_meta::generate_test_instances(list<rgw_bucket_dir_entry_meta*>& o)
-{
-  rgw_bucket_dir_entry_meta *m = new rgw_bucket_dir_entry_meta;
-  m->category = 1;
-  m->size = 100;
-  m->etag = "etag";
-  m->owner = "owner";
-  m->owner_display_name = "display name";
-  m->tag = "tag";
-  m->content_type = "content/type";
-  o.push_back(m);
-  o.push_back(new rgw_bucket_dir_entry_meta);
-}
-
-void rgw_bucket_dir_entry_meta::dump(Formatter *f) const
-{
-  f->dump_int("category", category);
-  f->dump_unsigned("size", size);
-  f->dump_stream("mtime") << mtime;
-  f->dump_string("etag", etag);
-  f->dump_string("owner", owner);
-  f->dump_string("owner_display_name", owner_display_name);
-  f->dump_string("tag", tag);
-  f->dump_string("content_type", content_type);
-}
-
-void rgw_bucket_dir_entry::generate_test_instances(list<rgw_bucket_dir_entry*>& o)
-{
-  list<rgw_bucket_dir_entry_meta *> l;
-  rgw_bucket_dir_entry_meta::generate_test_instances(l);
-
-  list<rgw_bucket_dir_entry_meta *>::iterator iter;
-  for (iter = l.begin(); iter != l.end(); ++iter) {
-    rgw_bucket_dir_entry_meta *m = *iter;
-    rgw_bucket_dir_entry *e = new rgw_bucket_dir_entry;
-    e->name = "name";
-    e->epoch = 1234;
-    e->locator = "locator";
-    e->exists = true;
-    e->meta = *m;
-
-    o.push_back(e);
-
-    delete m;
-  }
-  o.push_back(new rgw_bucket_dir_entry);
-}
-
-void rgw_bucket_dir_entry::dump(Formatter *f) const
-{
-  f->dump_string("name", name);
-  f->dump_unsigned("epoch", epoch);
-  f->dump_string("locator", locator);
-  f->dump_int("exists", (int)exists);
-  f->open_object_section("meta");
-  meta.dump(f);
-  f->close_section();
-
-  map<string, struct rgw_bucket_pending_info>::const_iterator iter = pending_map.begin();
-  f->open_array_section("pending_map");
-  for (; iter != pending_map.end(); ++iter) {
-    f->dump_string("tag", iter->first);
-    f->open_object_section("info");
-    iter->second.dump(f);
-    f->close_section();
-  }
-  f->close_section();
-}
-
-void rgw_bucket_category_stats::generate_test_instances(list<rgw_bucket_category_stats*>& o)
-{
-  rgw_bucket_category_stats *s = new rgw_bucket_category_stats;
-  s->total_size = 1024;
-  s->total_size_rounded = 4096;
-  s->num_entries = 2;
-  o.push_back(s);
-  o.push_back(new rgw_bucket_category_stats);
-}
-
-void rgw_bucket_category_stats::dump(Formatter *f) const
-{
-  f->dump_unsigned("total_size", total_size);
-  f->dump_unsigned("total_size_rounded", total_size_rounded);
-  f->dump_unsigned("num_entries", num_entries);
-}
-
-void rgw_bucket_dir_header::generate_test_instances(list<rgw_bucket_dir_header*>& o)
-{
-  list<rgw_bucket_category_stats *> l;
-  list<rgw_bucket_category_stats *>::iterator iter;
-  rgw_bucket_category_stats::generate_test_instances(l);
-
-  uint8_t i;
-  for (i = 0, iter = l.begin(); iter != l.end(); ++iter, ++i) {
-    rgw_bucket_dir_header *h = new rgw_bucket_dir_header;
-    rgw_bucket_category_stats *s = *iter;
-    h->stats[i] = *s;
-
-    o.push_back(h);
-
-    delete s;
-  }
-
-  o.push_back(new rgw_bucket_dir_header);
-}
-
-void rgw_bucket_dir_header::dump(Formatter *f) const
-{
-  map<uint8_t, struct rgw_bucket_category_stats>::const_iterator iter = stats.begin();
-  f->open_array_section("stats");
-  for (; iter != stats.end(); ++iter) {
-    f->dump_int("category", (int)iter->first);
-    f->open_object_section("category_stats");
-    iter->second.dump(f);
-    f->close_section();
-  }
-  f->close_section();
-}
-
-void rgw_bucket_dir::generate_test_instances(list<rgw_bucket_dir*>& o)
-{
-  list<rgw_bucket_dir_header *> l;
-  list<rgw_bucket_dir_header *>::iterator iter;
-  rgw_bucket_dir_header::generate_test_instances(l);
-
-  uint8_t i;
-  for (i = 0, iter = l.begin(); iter != l.end(); ++iter, ++i) {
-    rgw_bucket_dir *d = new rgw_bucket_dir;
-    rgw_bucket_dir_header *h = *iter;
-    d->header = *h;
-
-    list<rgw_bucket_dir_entry *> el;
-    list<rgw_bucket_dir_entry *>::iterator eiter;
-    for (eiter = el.begin(); eiter != el.end(); ++eiter) {
-      rgw_bucket_dir_entry *e = *eiter;
-      d->m[e->name] = *e;
-
-      delete e;
-    }
-
-    o.push_back(d);
-
-    delete h;
-  }
-
-  o.push_back(new rgw_bucket_dir);
-}
-
-void rgw_bucket_dir::dump(Formatter *f) const
-{
-  f->open_object_section("header");
-  header.dump(f);
-  f->close_section();
-  map<string, struct rgw_bucket_dir_entry>::const_iterator iter = m.begin();
-  f->open_array_section("map");
-  for (; iter != m.end(); ++iter) {
-    f->dump_string("obj", iter->first);
-    f->open_object_section("dir_entry");
-    iter->second.dump(f);
-    f->close_section();
-  }
-  f->close_section();
-}
-
-void rgw_cls_obj_prepare_op::generate_test_instances(list<rgw_cls_obj_prepare_op*>& o)
-{
-  rgw_cls_obj_prepare_op *op = new rgw_cls_obj_prepare_op;
-  op->op = CLS_RGW_OP_ADD;
-  op->name = "name";
-  op->tag = "tag";
-  op->locator = "locator";
-  o.push_back(op);
-  o.push_back(new rgw_cls_obj_prepare_op);
-}
-
-void rgw_cls_obj_prepare_op::dump(Formatter *f) const
-{
-  f->dump_int("op", op);
-  f->dump_string("name", name);
-  f->dump_string("tag", tag);
-  f->dump_string("locator", locator);
-}
-
-void rgw_cls_obj_complete_op::generate_test_instances(list<rgw_cls_obj_complete_op*>& o)
-{
-  rgw_cls_obj_complete_op *op = new rgw_cls_obj_complete_op;
-  op->op = CLS_RGW_OP_DEL;
-  op->name = "name";
-  op->locator = "locator";
-  op->epoch = 100;
-  op->tag = "tag";
-
-  list<rgw_bucket_dir_entry_meta *> l;
-  rgw_bucket_dir_entry_meta::generate_test_instances(l);
-  list<rgw_bucket_dir_entry_meta *>::iterator iter = l.begin();
-  op->meta = *(*iter);
-
-  o.push_back(op);
-
-  o.push_back(new rgw_cls_obj_complete_op);
-}
-
-void rgw_cls_obj_complete_op::dump(Formatter *f) const
-{
-  f->dump_int("op", (int)op);
-  f->dump_string("name", name);
-  f->dump_string("locator", locator);
-  f->dump_unsigned("epoch", epoch);
-  f->open_object_section("meta");
-  meta.dump(f);
-  f->close_section();
-  f->dump_string("tag", tag);
-}
-
-void rgw_cls_list_op::generate_test_instances(list<rgw_cls_list_op*>& o)
-{
-  rgw_cls_list_op *op = new rgw_cls_list_op;
-  op->start_obj = "start_obj";
-  op->num_entries = 100;
-  op->filter_prefix = "filter_prefix";
-  o.push_back(op);
-  o.push_back(new rgw_cls_list_op);
-}
-
-void rgw_cls_list_op::dump(Formatter *f) const
-{
-  f->dump_string("start_obj", start_obj);
-  f->dump_unsigned("num_entries", num_entries);
-}
-
-void rgw_cls_list_ret::generate_test_instances(list<rgw_cls_list_ret*>& o)
-{
- list<rgw_bucket_dir *> l;
-  rgw_bucket_dir::generate_test_instances(l);
-  list<rgw_bucket_dir *>::iterator iter;
-  for (iter = l.begin(); iter != l.end(); ++iter) {
-    rgw_bucket_dir *d = *iter;
-
-    rgw_cls_list_ret *ret = new rgw_cls_list_ret;
-    ret->dir = *d;
-    ret->is_truncated = true;
-
-    o.push_back(ret);
-
-    delete d;
-  }
-
-  o.push_back(new rgw_cls_list_ret);
-}
-
-void rgw_cls_list_ret::dump(Formatter *f) const
-{
-  f->open_object_section("dir");
-  dir.dump(f);
-  f->close_section();
-  f->dump_int("is_truncated", (int)is_truncated);
-}
 
+#define dout_subsys ceph_subsys_rgw
 
 
 class RGWWatcher : public librados::WatchCtx {
@@ -336,78 +67,6 @@ public:
   }
 };
 
-static void prepend_bucket_marker(rgw_bucket& bucket, string& orig_oid, string& oid)
-{
-  if (bucket.marker.empty() || orig_oid.empty()) {
-    oid = orig_oid;
-  } else {
-    oid = bucket.marker;
-    oid.append("_");
-    oid.append(orig_oid);
-  }
-}
-
-static void get_obj_bucket_and_oid_key(rgw_obj& obj, rgw_bucket& bucket, string& oid, string& key)
-{
-  bucket = obj.bucket;
-  prepend_bucket_marker(bucket, obj.object, oid);
-  prepend_bucket_marker(bucket, obj.key, key);
-}
-
-
-#define DOUT_SUBSYS rgw
-
-void RGWObjManifestPart::generate_test_instances(std::list<RGWObjManifestPart*>& o)
-{
-  o.push_back(new RGWObjManifestPart);
-
-  RGWObjManifestPart *p = new RGWObjManifestPart;
-  rgw_bucket b("bucket", ".pool", "marker_", "12");
-  p->loc = rgw_obj(b, "object");
-  p->loc_ofs = 512 * 1024;
-  p->size = 128 * 1024;
-  o.push_back(p);
-}
-
-void RGWObjManifestPart::dump(Formatter *f) const
-{
-  f->open_object_section("loc");
-  loc.dump(f);
-  f->close_section();
-  f->dump_unsigned("loc_ofs", loc_ofs);
-  f->dump_unsigned("size", size);
-}
-
-void RGWObjManifest::generate_test_instances(std::list<RGWObjManifest*>& o)
-{
-  RGWObjManifest *m = new RGWObjManifest;
-  for (int i = 0; i<10; i++) {
-    RGWObjManifestPart p;
-    rgw_bucket b("bucket", ".pool", "marker_", "12");
-    p.loc = rgw_obj(b, "object");
-    p.loc_ofs = 0;
-    p.size = 512 * 1024;
-    m->objs[(uint64_t)i * 512 * 1024] = p;
-  }
-  m->obj_size = 5 * 1024 * 1024;
-
-  o.push_back(new RGWObjManifest);
-}
-
-void RGWObjManifest::dump(Formatter *f) const
-{
-  map<uint64_t, RGWObjManifestPart>::const_iterator iter = objs.begin();
-  f->open_array_section("objs");
-  for (; iter != objs.end(); ++iter) {
-    f->dump_unsigned("ofs", iter->first);
-    f->open_object_section("part");
-    iter->second.dump(f);
-    f->close_section();
-  }
-  f->close_section();
-  f->dump_unsigned("obj_size", obj_size);
-}
-
 RGWRados *RGWRados::init_storage_provider(CephContext *cct)
 {
   int use_cache = cct->_conf->rgw_cache_enabled;
@@ -462,7 +121,20 @@ int RGWRados::initialize()
 
 void RGWRados::finalize_watch()
 {
-  control_pool_ctx.unwatch(notify_oid, watch_handle);
+  for (int i = 0; i < num_watchers; i++) {
+    string& notify_oid = notify_oids[i];
+    if (notify_oid.empty())
+      continue;
+    uint64_t watch_handle = watch_handles[i];
+    control_pool_ctx.unwatch(notify_oid, watch_handle);
+
+    RGWWatcher *watcher = watchers[i];
+    delete watcher;
+  }
+
+  delete[] notify_oids;
+  delete[] watch_handles;
+  delete[] watchers;
 }
 
 /**
@@ -500,14 +172,50 @@ int RGWRados::init_watch()
       return r;
   }
 
-  r = control_pool_ctx.create(notify_oid, false);
-  if (r < 0 && r != -EEXIST)
-    return r;
+  num_watchers = cct->_conf->rgw_num_control_oids;
 
-  watcher = new RGWWatcher(this);
-  r = control_pool_ctx.watch(notify_oid, 0, &watch_handle, watcher);
+  bool compat_oid = (num_watchers == 0);
 
-  return r;
+  if (num_watchers <= 0)
+    num_watchers = 1;
+
+  notify_oids = new string[num_watchers];
+  watchers = new RGWWatcher *[num_watchers];
+  watch_handles = new uint64_t[num_watchers];
+
+  for (int i=0; i < num_watchers; i++) {
+    string& notify_oid = notify_oids[i];
+    notify_oid = notify_oid_prefix;
+    if (!compat_oid) {
+      char buf[16];
+      snprintf(buf, sizeof(buf), ".%d", i);
+      notify_oid.append(buf);
+    }
+    r = control_pool_ctx.create(notify_oid, false);
+    if (r < 0 && r != -EEXIST)
+      return r;
+
+    RGWWatcher *watcher = new RGWWatcher(this);
+    watchers[i] = watcher;
+
+    r = control_pool_ctx.watch(notify_oid, 0, &watch_handles[i], watcher);
+    if (r < 0)
+      return r;
+  }
+
+  return 0;
+}
+
+void RGWRados::pick_control_oid(const string& key, string& notify_oid)
+{
+  uint32_t r = ceph_str_hash_linux(key.c_str(), key.size());
+
+  int i = r % num_watchers;
+  char buf[16];
+  snprintf(buf, sizeof(buf), ".%d", i);
+
+  notify_oid = notify_oid_prefix;
+  notify_oid.append(buf);
 }
 
 int RGWRados::open_bucket_ctx(rgw_bucket& bucket, librados::IoCtx&  io_ctx)
@@ -677,6 +385,144 @@ int RGWRados::log_show_next(RGWAccessHandle handle, rgw_log_entry *entry)
   return 1;
 }
 
+/**
+ * usage_log_hash: get usage log key hash, based on name and index
+ *
+ * Get the usage object name. Since a user may have more than 1
+ * object holding that info (multiple shards), we use index to
+ * specify that shard number. Once index exceeds max shards it
+ * wraps.
+ * If name is not being set, results for all users will be returned
+ * and index will wrap only after total shards number.
+ *
+ * @param cct [in] ceph context
+ * @param name [in] user name
+ * @param hash [out] hash value
+ * @param index [in] shard index number 
+ */
+static void usage_log_hash(CephContext *cct, const string& name, string& hash, uint32_t index)
+{
+  uint32_t val = index;
+
+  if (!name.empty()) {
+    int max_user_shards = max(cct->_conf->rgw_usage_max_user_shards, 1);
+    val %= max_user_shards;
+    val += ceph_str_hash_linux(name.c_str(), name.size());
+  }
+  char buf[16];
+  int max_shards = max(cct->_conf->rgw_usage_max_shards, 1);
+  snprintf(buf, sizeof(buf), RGW_USAGE_OBJ_PREFIX "%u", (unsigned)(val % max_shards));
+  hash = buf;
+}
+
+int RGWRados::log_usage(map<rgw_user_bucket, RGWUsageBatch>& usage_info)
+{
+  uint32_t index = 0;
+
+  map<string, rgw_usage_log_info> log_objs;
+
+  string hash;
+  string last_user;
+
+  /* restructure usage map, cluster by object hash */
+  map<rgw_user_bucket, RGWUsageBatch>::iterator iter;
+  for (iter = usage_info.begin(); iter != usage_info.end(); ++iter) {
+    const rgw_user_bucket& ub = iter->first;
+    RGWUsageBatch& info = iter->second;
+
+    if (ub.user.empty()) {
+      ldout(cct, 0) << "WARNING: RGWRados::log_usage(): user name empty (bucket=" << ub.bucket << "), skipping" << dendl;
+      continue;
+    }
+
+    if (ub.user != last_user) {
+      /* index *should* be random, but why waste extra cycles
+         in most cases max user shards is not going to exceed 1,
+         so just incrementing it */
+      usage_log_hash(cct, ub.user, hash, index++);
+    }
+    last_user = ub.user;
+    vector<rgw_usage_log_entry>& v = log_objs[hash].entries;
+
+    map<utime_t, rgw_usage_log_entry>::iterator miter;
+    for (miter = info.m.begin(); miter != info.m.end(); ++miter) {
+      v.push_back(miter->second);
+    }
+  }
+
+  map<string, rgw_usage_log_info>::iterator liter;
+
+  for (liter = log_objs.begin(); liter != log_objs.end(); ++liter) {
+    int r = cls_obj_usage_log_add(liter->first, liter->second);
+    if (r < 0)
+      return r;
+  }
+  return 0;
+}
+
+int RGWRados::read_usage(string& user, uint64_t start_epoch, uint64_t end_epoch, uint32_t max_entries,
+                         bool *is_truncated, RGWUsageIter& usage_iter, map<rgw_user_bucket, rgw_usage_log_entry>& usage)
+{
+  uint32_t num = max_entries;
+  string hash, first_hash;
+  usage_log_hash(cct, user, first_hash, 0);
+
+  if (usage_iter.index) {
+    usage_log_hash(cct, user, hash, usage_iter.index);
+  } else {
+    hash = first_hash;
+  }
+
+  do {
+    rgw_cls_usage_log_read_ret read_ret;
+    map<rgw_user_bucket, rgw_usage_log_entry>::iterator iter;
+
+    int ret =  cls_obj_usage_log_read(hash, user, start_epoch, end_epoch, num,
+                                    usage_iter.read_iter, read_ret, is_truncated);
+    if (ret == -ENOENT)
+      goto next;
+
+    if (ret < 0)
+      return ret;
+
+    num -= usage.size();
+
+    for (iter = read_ret.usage.begin(); iter != read_ret.usage.end(); ++iter) {
+      usage[iter->first].aggregate(iter->second);
+    }
+
+next:
+    if (!*is_truncated) {
+      usage_iter.read_iter.clear();
+      usage_log_hash(cct, user, hash, ++usage_iter.index);
+    }
+  } while (num && !*is_truncated && hash != first_hash);
+  return 0;
+}
+
+int RGWRados::trim_usage(string& user, uint64_t start_epoch, uint64_t end_epoch)
+{
+  uint32_t index = 0;
+  string hash, first_hash;
+  usage_log_hash(cct, user, first_hash, index);
+
+  hash = first_hash;
+
+  do {
+    int ret =  cls_obj_usage_log_trim(hash, user, start_epoch, end_epoch);
+    if (ret == -ENOENT)
+      goto next;
+
+    if (ret < 0)
+      return ret;
+
+next:
+    usage_log_hash(cct, user, hash, ++index);
+  } while (hash != first_hash);
+
+  return 0;
+}
+
 int RGWRados::decode_policy(bufferlist& bl, ACLOwner *owner)
 {
   bufferlist::iterator i = bl.begin();
@@ -851,12 +697,6 @@ int RGWRados::store_bucket_info(RGWBucketInfo& info, map<string, bufferlist> *pa
   if (ret < 0)
     return ret;
 
-  ret = rgw_put_obj(info.owner, pi_buckets_rados, info.bucket.bucket_id, bl.c_str(), bl.length(), false, pattrs);
-  if (ret < 0) {
-    ldout(cct, 0) << "ERROR: failed to store " << pi_buckets_rados << ":" << info.bucket.bucket_id << " ret=" << ret << dendl;
-    return ret;
-  }
-
   ldout(cct, 20) << "store_bucket_info: bucket=" << info.bucket << " owner " << info.owner << dendl;
   return 0;
 }
@@ -1247,13 +1087,9 @@ bool RGWRados::aio_completed(void *handle)
 }
 /**
  * Copy an object.
- * dest_bucket: the bucket to copy into
  * dest_obj: the object to copy into
- * src_bucket: the bucket to copy from
  * src_obj: the object to copy from
- * mod_ptr, unmod_ptr, if_match, if_nomatch: as used in get_obj
- * attrs: these are placed on the new object IN ADDITION to
- *    (or overwriting) any attrs copied from the original object
+ * attrs: if replace_attrs is set then these are placed on the new object
  * err: stores any errors resulting from the get of the original object
  * Returns: 0 on success, -ERR# otherwise.
  */
@@ -1265,7 +1101,8 @@ int RGWRados::copy_obj(void *ctx,
                const time_t *unmod_ptr,
                const char *if_match,
                const char *if_nomatch,
-               map<string, bufferlist>& attrs,  /* in/out */
+               bool replace_attrs,
+               map<string, bufferlist>& attrs,
                RGWObjCategory category,
                struct rgw_err *err)
 {
@@ -1337,12 +1174,14 @@ int RGWRados::copy_obj(void *ctx,
   }
   manifest.obj_size = ofs;
 
-  for (iter = attrs.begin(); iter != attrs.end(); ++iter) {
-    attrset[iter->first] = iter->second;
-  }
-  attrs = attrset;
+  if (replace_attrs) {
+    if (!attrs[RGW_ATTR_ETAG].length())
+      attrs[RGW_ATTR_ETAG] = attrset[RGW_ATTR_ETAG];
 
-  ret = rgwstore->put_obj_meta(ctx, dest_obj, end + 1, NULL, attrs, category, false, NULL, &first_chunk, &manifest);
+    attrset = attrs;
+  }
+
+  ret = rgwstore->put_obj_meta(ctx, dest_obj, end + 1, NULL, attrset, category, false, NULL, &first_chunk, &manifest);
   if (mtime)
     obj_stat(ctx, dest_obj, NULL, mtime, NULL, NULL);
 
@@ -1421,7 +1260,8 @@ int RGWRados::set_buckets_enabled(vector<rgw_bucket>& buckets, bool enabled)
       ldout(cct, 20) << "disabling bucket name=" << bucket.name << dendl;
 
     RGWBucketInfo info;
-    int r = get_bucket_info(NULL, bucket.name, info);
+    map<string, bufferlist> attrs;
+    int r = get_bucket_info(NULL, bucket.name, info, &attrs);
     if (r < 0) {
       ldout(cct, 0) << "NOTICE: get_bucket_info on bucket=" << bucket.name << " returned err=" << r << ", skipping bucket" << dendl;
       ret = r;
@@ -1433,7 +1273,7 @@ int RGWRados::set_buckets_enabled(vector<rgw_bucket>& buckets, bool enabled)
       info.flags |= BUCKET_SUSPENDED;
     }
 
-    r = put_bucket_info(bucket.name, info, false);
+    r = put_bucket_info(bucket.name, info, false, &attrs);
     if (r < 0) {
       ldout(cct, 0) << "NOTICE: put_bucket_info on bucket=" << bucket.name << " returned err=" << r << ", skipping bucket" << dendl;
       ret = r;
@@ -1976,7 +1816,7 @@ int RGWRados::prepare_get_obj(void *ctx, rgw_obj& obj,
 
   if (attrs) {
     *attrs = astate->attrset;
-    if (cct->_conf->debug_rgw >= 20) {
+    if (cct->_conf->subsys.should_gather(ceph_subsys_rgw, 20)) {
       for (iter = attrs->begin(); iter != attrs->end(); ++iter) {
         ldout(cct, 20) << "Read xattr: " << iter->first << dendl;
       }
@@ -2473,11 +2313,11 @@ int RGWRados::get_bucket_stats(rgw_bucket& bucket, map<RGWObjCategory, RGWBucket
   return 0;
 }
 
-int RGWRados::get_bucket_info(void *ctx, string& bucket_name, RGWBucketInfo& info)
+int RGWRados::get_bucket_info(void *ctx, string& bucket_name, RGWBucketInfo& info, map<string, bufferlist> *pattrs)
 {
   bufferlist bl;
 
-  int ret = rgw_get_obj(ctx, pi_buckets_rados, bucket_name, bl);
+  int ret = rgw_get_obj(ctx, pi_buckets_rados, bucket_name, bl, pattrs);
   if (ret < 0) {
     if (ret != -ENOENT)
       return ret;
@@ -2500,7 +2340,7 @@ int RGWRados::get_bucket_info(void *ctx, string& bucket_name, RGWBucketInfo& inf
   return 0;
 }
 
-int RGWRados::put_bucket_info(string& bucket_name, RGWBucketInfo& info, bool exclusive)
+int RGWRados::put_bucket_info(string& bucket_name, RGWBucketInfo& info, bool exclusive, map<string, bufferlist> *pattrs)
 {
   bufferlist bl;
 
@@ -2508,7 +2348,7 @@ int RGWRados::put_bucket_info(string& bucket_name, RGWBucketInfo& info, bool exc
 
   string unused;
 
-  int ret = rgw_put_obj(unused, pi_buckets_rados, bucket_name, bl.c_str(), bl.length(), exclusive);
+  int ret = rgw_put_obj(unused, pi_buckets_rados, bucket_name, bl.c_str(), bl.length(), exclusive, pattrs);
 
   return ret;
 }
@@ -2642,8 +2482,11 @@ int RGWRados::append_async(rgw_obj& obj, size_t size, bufferlist& bl)
   return r;
 }
 
-int RGWRados::distribute(bufferlist& bl)
+int RGWRados::distribute(const string& key, bufferlist& bl)
 {
+  string notify_oid;
+  pick_control_oid(key, notify_oid);
+
   ldout(cct, 10) << "distributing notification oid=" << notify_oid << " bl.length()=" << bl.length() << dendl;
   int r = control_pool_ctx.notify(notify_oid, 0, bl);
   return r;
@@ -2846,6 +2689,12 @@ int RGWRados::cls_bucket_list(rgw_bucket& bucket, string start, string prefix,
     e.owner_display_name = dirent.meta.owner_display_name;
     e.content_type = dirent.meta.content_type;
 
+    /* oh, that shouldn't happen! */
+    if (e.name.empty()) {
+      ldout(cct, 0) << "WARNING: got empty dirent name, skipping" << dendl;
+      continue;
+    }
+
     if (!dirent.exists || !dirent.pending_map.empty()) {
       /* there are uncommitted ops. We need to check the current state,
        * and if the tags are old we need to do cleanup as well. */
@@ -2870,12 +2719,93 @@ int RGWRados::cls_bucket_list(rgw_bucket& bucket, string start, string prefix,
   if (updates.length()) {
     // we don't care if we lose suggested updates, send them off blindly
     AioCompletion *c = librados::Rados::aio_create_completion(NULL, NULL, NULL);
-    r = io_ctx.aio_exec(oid, c, "rgw", "dir_suggest_changes", in, NULL);
+    r = io_ctx.aio_exec(oid, c, "rgw", "dir_suggest_changes", updates, NULL);
     c->release();
   }
   return m.size();
 }
 
+int RGWRados::cls_obj_usage_log_add(const string& oid, rgw_usage_log_info& info)
+{
+  librados::IoCtx io_ctx;
+
+  int r = rados->ioctx_create(RGW_USAGE_LOG_POOL_NAME, io_ctx);
+  if (r == -ENOENT) {
+    string id;
+    map<std::string, bufferlist> attrs;
+    rgw_bucket pool(RGW_USAGE_LOG_POOL_NAME);
+    r = rgwstore->create_bucket(id, pool, attrs, true);
+    if (r < 0)
+      return r;
+ 
+    // retry
+    r = rados->ioctx_create(RGW_USAGE_LOG_POOL_NAME, io_ctx);
+  }
+  if (r < 0)
+    return r;
+
+  bufferlist in, out;
+  rgw_cls_usage_log_add_op call;
+  call.info = info;
+  ::encode(call, in);
+  r = io_ctx.exec(oid, "rgw", "user_usage_log_add", in, out);
+  return r;
+}
+
+int RGWRados::cls_obj_usage_log_read(string& oid, string& user, uint64_t start_epoch, uint64_t end_epoch, uint32_t max_entries,
+                                     string& read_iter, rgw_cls_usage_log_read_ret& result, bool *is_truncated)
+{
+  librados::IoCtx io_ctx;
+
+  *is_truncated = false;
+
+  int r = rados->ioctx_create(RGW_USAGE_LOG_POOL_NAME, io_ctx);
+  if (r < 0)
+    return r;
+
+  bufferlist in, out;
+  rgw_cls_usage_log_read_op call;
+  call.start_epoch = start_epoch;
+  call.end_epoch = end_epoch;
+  call.owner = user;
+  call.max_entries = max_entries;
+  call.iter = read_iter;
+  ::encode(call, in);
+  r = io_ctx.exec(oid, "rgw", "user_usage_log_read", in, out);
+  if (r < 0)
+    return r;
+
+  try {
+    bufferlist::iterator iter = out.begin();
+    ::decode(result, iter);
+    read_iter = result.next_iter;
+    if (is_truncated)
+      *is_truncated = result.truncated;
+  } catch (buffer::error& e) {
+    ldout(cct, 0) << "ERROR: cls_obj_usage_log_read_op: failed to decode result" << dendl;
+    return -EINVAL;
+  }
+  return r;
+}
+
+int RGWRados::cls_obj_usage_log_trim(string& oid, string& user, uint64_t start_epoch, uint64_t end_epoch)
+{
+  librados::IoCtx io_ctx;
+
+  int r = rados->ioctx_create(RGW_USAGE_LOG_POOL_NAME, io_ctx);
+  if (r < 0)
+    return r;
+
+  bufferlist in, out;
+  rgw_cls_usage_log_trim_op call;
+  call.user = user;
+  call.start_epoch = start_epoch;
+  call.end_epoch = end_epoch;
+  ::encode(call, in);
+  r = io_ctx.exec(oid, "rgw", "user_usage_log_trim", in, out);
+  return r;
+}
+
 int RGWRados::check_disk_state(librados::IoCtx io_ctx,
                                rgw_bucket& bucket,
                                rgw_bucket_dir_entry& list_state,
