@@ -253,9 +253,13 @@ void dump_continue(struct req_state *s)
   FCGX_FFlush(s->fcgx->out);
 }
 
-void dump_range(struct req_state *s, off_t ofs, off_t end, size_t total)
+void dump_range(struct req_state *s, uint64_t ofs, uint64_t end, uint64_t total)
 {
-    CGI_PRINTF(s,"Content-Range: bytes %d-%d/%d\n", (int)ofs, (int)end, (int)total);
+  char range_buf[128];
+
+  /* dumping range into temp buffer first, as libfcgi will fail to digest %lld */
+  snprintf(range_buf, sizeof(range_buf), "%lld-%lld/%lld", (long long)ofs, (long long)end, (long long)total);
+  CGI_PRINTF(s,"Content-Range: bytes %s\n", range_buf);
 }
 
 int RGWGetObj_REST::get_params()
@@ -347,6 +351,47 @@ int RGWInitMultipart_REST::get_params()
   return ret;
 }
 
+static int read_all_chunked_input(req_state *s, char **pdata, int *plen)
+{
+#define READ_CHUNK 4096
+#define MAX_READ_CHUNK (128 * 1024)
+  int need_to_read = READ_CHUNK;
+  int total = need_to_read;
+  char *data = (char *)malloc(total + 1);
+  if (!data)
+    return -ENOMEM;
+
+  int read_len = 0, len = 0;
+  do {
+    CGI_GetStr(s, data + len, need_to_read, read_len);
+
+    len += read_len;
+
+    if (read_len == need_to_read) {
+      if (need_to_read < MAX_READ_CHUNK)
+	need_to_read *= 2;
+
+      total += need_to_read;
+
+      void *p = realloc(data, total + 1);
+      if (!p) {
+        free(data);
+        return -ENOMEM;
+      }
+      data = (char *)p;
+    } else {
+      break;
+    }
+
+  } while (true);
+  data[len] = '\0';
+
+  *pdata = data;
+  *plen = len;
+
+  return 0;
+}
+
 int RGWCompleteMultipart_REST::get_params()
 {
   upload_id = s->args.get("uploadId");
@@ -355,7 +400,6 @@ int RGWCompleteMultipart_REST::get_params()
     ret = -ENOTSUP;
     return ret;
   }
-
   size_t cl = 0;
 
   if (s->length)
@@ -369,7 +413,13 @@ int RGWCompleteMultipart_REST::get_params()
     CGI_GetStr(s, data, cl, len);
     data[len] = '\0';
   } else {
-    len = 0;
+    const char *encoding = s->env->get("HTTP_TRANSFER_ENCODING");
+    if (!encoding || strcmp(encoding, "chunked") != 0)
+      return -ERR_LENGTH_REQUIRED;
+
+    ret = read_all_chunked_input(s, &data, &len);
+    if (ret < 0)
+      return ret;
   }
 
   return ret;
